"""

File:    PbxConfigJobRunner.py
Author:  Jonathan Maurice <jmaurice@sangoma.com>
Date:    4/6/2011 5:16:50 PM
 
Description: Job runner of the pbx-configurator. This class is responsible
of calling the run method on each jobs it has. It's also used to notifiy
progress observer.
 
Copyright (C) 2003-2011 Sangoma Technologies. 
All Rights Reserved.

This code is Sangoma Technologies Confidential Property.
Use of and Access to this code is covered by a previously executed
Non-disclosure Agreement between Sangoma Technologies and the Recipient.
This code is being supplied for evaluation purposes only and is not to be
used for any other purpose.  

"""


from paraxip.infra.PyLoggingFunctions import *
from netborderPyInfra import TraceScope
from paraxip.infra.PyJob import JobRunner

import paraxip.pbx_configurator.PbxConfigJobs as PbxJobs 
from paraxip.smart_update.SmartUpdateJobRunner import SmartUpdateJobRunner

fsLogger = Logger("netborder.pbxConfigurator.PbxConfigJobRunner")


class PbxConfigJobRunner(SmartUpdateJobRunner):

  def __init__(self, jobStatus, pbxConfigMan):
    t = TraceScope(fsLogger, "PbxConfigJobRunner.__init__")
    super(PbxConfigJobRunner, self).__init__(jobStatus, pbxConfigMan)
    self._smartUpdateJob = PbxJobs.SmartUpdateJob(jobStatus, pbxConfigMan)

  
  def configure(self):
    """Add all needed job"""
    t = TraceScope(fsLogger, "PbxConfigJobRunner.configure")
    
    self.addJob(self._tempDirInitalJob)
    self.addJob(self._setTempModeJob)

    #This file is created by the PbxConfigManager.
    import PbxConfigJobAdder
    PbxConfigJobAdder.addJobs(self,
                              self._jobStatus,
                              self._tempDirManager,
                              self._swapJob,
                              self._smartUpdateJob,
                              self._setFinalModeJob)

    self.addJob(self._cleanDirJob)

    return True

    
  def run(self):
    """Call JobRunner default run implementation"""
    t = TraceScope(fsLogger, "PbxConfigJobRunner.run")
    return super(PbxConfigJobRunner, self).run()

## !!MIRROR FILE, DO NOT MODIFY!!
