"""

File: SmartUpdateJobRunner.py
Author: Jonathan Maurice
Date: 16/02/2011
Description: Job runner of the smart update process. This class is responsible
of calling the run method on each jobs it has. It's also used to notifiy
progress observer.

Copyright (C) 2003-2011 Sangoma Technologies.
All Rights Reserved.

"""

from paraxip.infra.PyLoggingFunctions import  *
from netborderPyInfra import TraceScope
from paraxip.infra.PyJob import JobRunner
import paraxip.smart_update.TempDirJobs as TmpJobs

fsLogger = Logger("netborder.smartUpdate.SmartUpdateJobRunner")


class SmartUpdateJobRunner(JobRunner):

  def __init__(self, jobStatus, tempDirManager):
    t = TraceScope(fsLogger, "SmartUpdateJobRunner.__init__")
    super(SmartUpdateJobRunner, self).__init__(jobStatus)
    self._tempDirManager = tempDirManager
    self._tempDirInitalJob = TmpJobs.TempDirInitialJob(self._tempDirManager)
    self._swapJob = TmpJobs.SwapJob(self._tempDirManager) 
    self._setTempModeJob = TmpJobs.SetTempModeJob(self._tempDirManager)
    self._setFinalModeJob = TmpJobs.SetFinalModeJob(self._tempDirManager)
    self._cleanDirJob = TmpJobs.CleanTempDirectoriesJob(self._tempDirManager)


  def configure(self):
    """Add all needed job"""
    t = TraceScope(fsLogger, "SmartUpdateJobRunner.configure")
    
    self.addJob(self._tempDirInitalJob)
    self.addJob(self._setTempModeJob)

    #This file is created by SmartUpdate.py
    import SmartUpdateAddJobs
    SmartUpdateAddJobs.addJobs(self,
                               self._jobStatus,
                               self._tempDirManager,
                               self._swapJob,
                               self._setFinalModeJob)

    self.addJob(self._cleanDirJob)
    return True


  def run(self):
    """Call JobRunner default run implementation"""
    t = TraceScope(fsLogger, "SmartUpdateJobRunner.run")
    return super(SmartUpdateJobRunner, self).run()

## !!MIRROR FILE, DO NOT MODIFY!!
