<?xml version="1.0" encoding="utf-8"?>

<!--

This stylesheet convert a gw.properties from a gateway version of 3.0.6 to
version 4.0.0. To do so:

-sip.contactHostName:
  If sip.contactHostName=automatic-hostname, comment and add 
  sip.contactHostNameMode=automatic-hostname.

  If sip.contactHostName=something else, add 
  sip.contactHostNameMode=manual and keep sip.contactHostName
  to the same value.

  If sip.contactHostName is missing, add sip.contactHostNameMode=automatic-dotted-ip.

-Add netborder.sangoma.fxotune.appPath and netborder.sangoma.fxotune.configPath

-Add Installation.Bin.Directory and Installation.Lib.Directory elements.

-->

<xsl:stylesheet version="1.0"
xmlns:xsl="http://www.w3.org/1999/XSL/Transform">
<xsl:output method = "xml" indent = "yes" /> 


<!-- This template copy all the elements and their attributes -->
<xsl:template match="node()|@*">
  <xsl:copy>
    <xsl:apply-templates select="node()|@*"/>
  </xsl:copy>
</xsl:template>

<!-- 
Add netborder.sangoma.fxotune.appPath and
netborder.sangoma.fxotune.configPath elements 
-->
<xsl:template match="netborder">
  <xsl:copy>
    <xsl:element name="sangoma">
      <xsl:element name="fxotune">
        <!-- fxotune.appPath -->
        <xsl:element name="appPath">
          <xsl:attribute name="value">
            <xsl:value-of select="//Installation/Directory/@value"/><xsl:text>/bin/netborder-fxo-tune.exe</xsl:text>
          </xsl:attribute>
          <xsl:attribute name="defined">true</xsl:attribute>
        </xsl:element>
        <!-- fxotune.configPath -->
        <xsl:element name="configPath">
          <xsl:attribute name="value">
            <xsl:value-of select="//Installation/Directory/@value"/><xsl:text>/config/fxo.conf</xsl:text>
          </xsl:attribute>
          <xsl:attribute name="defined">true</xsl:attribute>
        </xsl:element>
      </xsl:element>
    </xsl:element>
    <xsl:apply-templates select="node()|@*"/>
  </xsl:copy>
</xsl:template>


<!-- Add Installation.Bin.Directory and Installation.Lib.Directory elements -->
<xsl:template match="Installation">
  <xsl:copy>
    <xsl:element name="Bin">
      <xsl:element name="Directory">
        <xsl:attribute name="value">
          <xsl:value-of select="//Installation/Directory/@value"/><xsl:text>/bin</xsl:text>
        </xsl:attribute>
        <xsl:attribute name="defined">true</xsl:attribute>
      </xsl:element>
    </xsl:element>
    <xsl:element name="Lib">
      <xsl:element name="Directory">
        <xsl:attribute name="value">
           <xsl:value-of select="//Installation/Directory/@value"/><xsl:text>/bin</xsl:text>
        </xsl:attribute>
        <xsl:attribute name="defined">true</xsl:attribute>
      </xsl:element>
    </xsl:element>
    <xsl:apply-templates select="node()|@*"/>
  </xsl:copy>
</xsl:template>


<!-- 
Apply templates on all on all nodes before contactHostName. When
contactHostName is encountered do the operations stated above.
Afterward, apply-templates on all nodse following contactHostName.
-->
<xsl:template match="netborder/sip">
  <xsl:copy>
    <xsl:for-each select="*[following-sibling::contactHostName]">
        <xsl:copy>
          <xsl:apply-templates select="@*"/>
          <xsl:apply-templates/>
        </xsl:copy>
    </xsl:for-each>
    <xsl:apply-templates select="contactHostName"/>
    <xsl:choose>
      <xsl:when test="contactHostName[@defined='true']">
        <xsl:choose>
          <xsl:when test="contactHostName[@value='automatic-hostname']">
            <xsl:element name="contactHostNameMode">
              <xsl:attribute name="value">automatic-hostname</xsl:attribute>
              <xsl:attribute name="defined">true</xsl:attribute>
            </xsl:element>
          </xsl:when>
          <xsl:otherwise>
            <xsl:element name="contactHostNameMode">
              <xsl:attribute name="value">manual</xsl:attribute>
              <xsl:attribute name="defined">true</xsl:attribute>
            </xsl:element>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:when>
      <xsl:otherwise>
        <xsl:element name="contactHostNameMode">
          <xsl:attribute name="value">automatic-dotted-ip</xsl:attribute>
          <xsl:attribute name="defined">true</xsl:attribute>
        </xsl:element>
      </xsl:otherwise>
    </xsl:choose>
     <xsl:for-each select="*[preceding-sibling::contactHostName]">
      <xsl:copy>
        <xsl:apply-templates select="@*"/>
        <xsl:apply-templates/>
      </xsl:copy>
    </xsl:for-each>
  </xsl:copy>
</xsl:template>


<!-- Comment contactHostName if it's set to automatic-hostname  -->
<xsl:template match="contactHostName">
  <xsl:copy>
    <xsl:apply-templates select="@*"/>
    <xsl:if test="@value='automatic-hostname'">
      <xsl:attribute name="defined">false</xsl:attribute>
    </xsl:if>
  </xsl:copy>
</xsl:template>

</xsl:stylesheet> 

