import turbogears
from turbogears import controllers, expose
from turbogears import flash, redirect, url
from pstnconfig import model
from paraxip.gateway_gui import extjs
from operator import itemgetter
import simplejson
import os

import logging
log = logging.getLogger("pstnconfig.controllers")

class PhysicalConfigController(controllers.Controller):
#################################################
#                                               #
# Physical Config methods (Add/delete/get/save) #
#                                               #
#################################################
  ##########################################################################
  @expose()
  def addPhysicalConfigFromEditor (self, **data):
    """Adds a physical config from the editor"""
    log.info("Adding %s physical config" % data['configType'])
    if data['configType'] == 'T1':
      T1ConfigParams = {}
      T1ConfigParams['lineEncoding'] = 'B8ZS'
      T1ConfigParams['clocking'] = 'TERMINAL'
      T1ConfigParams['loopbackMode'] = 'DISABLED'
      T1ConfigParams['framing'] = 'ESF'
      T1ConfigParams['LBO'] = '0dB'

      #Create the audio parameters dictionary
      
      config = model.addT1SangomaDigitalConfig(model.currentConfig,
          'new_digital_T1_%s' % model.getNextPhysConfID(model.currentConfig),T1ConfigParams)
      
    elif data['configType'] == 'J1':
      J1ConfigParams = {}
      J1ConfigParams['lineEncoding'] = 'B8ZS'
      J1ConfigParams['clocking'] = 'TERMINAL'
      J1ConfigParams['loopbackMode'] = 'DISABLED'
      J1ConfigParams['framing'] = 'ESF'
      J1ConfigParams['LBO'] = '0dB'
      
      config = model.addJ1SangomaDigitalConfig(model.currentConfig,
          'new_digital_J1_%s' % model.getNextPhysConfID(model.currentConfig),J1ConfigParams)
      
    elif data['configType'] == 'E1':
      E1ConfigParams = {}
      E1ConfigParams['lineEncoding'] = 'HDB3'
      E1ConfigParams['E1Signaling'] = 'CCS'
      E1ConfigParams['clocking'] = 'TERMINAL'
      E1ConfigParams['loopbackMode'] = 'DISABLED'
      E1ConfigParams['framing'] = 'CRC4'
      E1ConfigParams['E1Termination'] = '120ohm'

      #Create the audio parameters dictionary
      E1AudioParams = model.defaultAudioParams.copy()
      E1AudioParams['defaultPcmLaw'] = 'PCMA'

      config = model.addE1SangomaDigitalConfig(model.currentConfig,
          'new_digital_E1_%s' % model.getNextPhysConfID(model.currentConfig),E1ConfigParams,E1AudioParams)
    
    elif data['configType'] == 'BRI':
      BRIConfigParams = {}
      BRIConfigParams['BRISignaling'] = 'CCS'
      BRIConfigParams['clocking'] = 'TERMINAL' # TEMP
      BRIConfigParams['loopbackMode'] = 'DISABLED'
      BRIConfigParams['LBO'] = '0dB'
      
      config = model.addBRISangomaDigitalConfig(model.currentConfig,
          'new_digital_BRI_%s' % model.getNextPhysConfID(model.currentConfig),BRIConfigParams)
       
    elif data['configType'] == 'Analog':
      config = \
        model.addSangomaAnalogConfig(
          model.currentConfig,
          'new_analog_%s' % model.getNextPhysConfID(model.currentConfig),
          'USA',
          "INTERNAL"
        )

    model.writeCurrentConfigToFile()

    return(str(extjs.FormActionSubmitResponse(success=True)))

  ##########################################################################
  @expose()
  def deletePhysicalConfigFromEditor (self, **data):
    """Deltes a physical config from the editor"""

    log.info("Deleting physical config %s" % data['physConfigs'])

    for selection in simplejson.loads(data['physConfigs']):
      config = model.getObjectFromEditorId(selection['nodeId'])
      if type(config) == model.sangomaDigitalConfig:
        model.currentConfig.get_physicalInterfaceConfigs().get_sangomaDigitalConfig().remove(config)
      elif type(config) == model.sangomaAnalogConfig:
        model.currentConfig.get_physicalInterfaceConfigs().get_sangomaAnalogConfig().remove(config)
      model.deleteObjectFromEditor(config)

    model.writeCurrentConfigToFile()

    return(str(extjs.FormActionSubmitResponse(success=True)))

  ##########################################################################
  # Returns a list of all the values associated to the selected T1 config.
  # The fields to be exposed to the user are:
  #   - registered id (hidden field).
  #   - The display name of the configs
  #   - DigitalT1 values
  #   - DChannelT1 values
  #   - Audio values including EchoCancellation, AcousticEchoCancellation, etc.
  @expose() # Ext.form.Action.load - expects JSON
  def getFrmDigitalT1PhysConfData(self, id):
    config = model.getObjectFromEditorId(id)
    digital = config.get_digitalT1()
    dchannel = digital.get_DChannelT1()
    audio = config.get_audio()
    data = {
      't1NodeId': id,
      'id_textT1': config.get_ID(),
      'name_textT1': config.get_displayName(),
      #DigitalT1 values
      'framingT1_selection': digital.get_framing(),
      'line_encT1_selection': digital.get_lineEncoding(),
      'lboT1_selection': digital.get_LBO(),
      'clockingT1_selection': digital.get_clocking(),
      'loopback_modeT1_selection': digital.get_loopbackMode(),
      #Audio values
      'defaultPcmLawT1_selection': audio.get_defaultPcmLaw(),
      'voicePacketLengthInMsT1_text':audio.get_voicePacketLengthInMs(),
      'toneSetAttributesT1_select':audio.get_toneSetName(),
      'optimalTxBufferLevelT1_text':audio.get_optimalBchannelTxQueueLevel(),
      #EchoCancellation values
      'ecT1_type_selection':audio.get_EchoCancellation().get_type(),
      'ecT1_mode_selection':audio.get_EchoCancellation().get_mode(),
      'ecT1_comfortNoiseMode_selection':audio.get_EchoCancellation().get_comfortNoiseMode(),
      'ecT1_tailDisplacementInMs_text':audio.get_EchoCancellation().get_tailDisplacementInMs(),
      'ecT1_doubleTalkBehavior_selection':audio.get_EchoCancellation().get_doubleTalkBehavior(),
      #Acoustic values
      'aecT1_type_selection':audio.get_AcousticEchoCancellation().get_type(),
      'aecT1_mode_selection':audio.get_AcousticEchoCancellation().get_mode(),
      #LevelControl values
      'levelT1_type_selection':audio.get_LevelControl().get_type(),
      'levelT1_direction_selection':audio.get_LevelControl().get_direction(),
      'levelT1_mode_selection':audio.get_LevelControl().get_mode(),
      'levelT1_targetLevelIndBM0_text':audio.get_LevelControl().get_targetLevelIndBM0(),
      'levelT1_gainIndB_text':audio.get_LevelControl().get_gainIndB(),
      #BackgroundNoiseAttenuation values
      'bnaT1_type_selection':audio.get_BackgroundNoiseAttenuation().get_type(),
      'bnaT1_direction_selection':audio.get_BackgroundNoiseAttenuation().get_direction(),
      'bnaT1_mode_selection':audio.get_BackgroundNoiseAttenuation().get_mode(),
      'bnaT1_attenuationIndB_selection':audio.get_BackgroundNoiseAttenuation().get_attenuationIndB(),
      #DTMFRemoval values
      'dtmfT1_type_selection':audio.get_DTMFRemoval().get_type(),
      'dtmfT1_direction_selection':audio.get_DTMFRemoval().get_direction(),
      'dtmfT1_mode_selection':audio.get_DTMFRemoval().get_mode()
    }
    #DChannelT1 values
    if dchannel is None:
      data['disableDChannelT1'] = 'DISABLED'
    else:
      data['disableDChannelT1'] = 'ENABLED'
      data['channelT1_idx_text'] = dchannel.get_channelIndex()
      data['mtu_sizeT1_text'] = dchannel.get_mtuSize()
      data['record_to_pcap_fileT1'] = dchannel.get_RecordToPcapFile()
      data['pcap_files_folderT1'] = dchannel.get_PcapFilesFolder()

    return(str(extjs.FormActionLoadResponse(success=True, data=data)))

  ##########################################################################
  # Returns a list of all the values associated to the selected E1 config.
  # The fields to be exposed to the user are:
  #   - registered id (hidden field).
  #   - The display name of the configs
  #   - DigitalE1 values
  #   - DChannelE1 values
  #   - Audio values including EchoCancellation, AcousticEchoCancellation, etc.
  @expose() # Ext.form.Action.load - expects JSON
  def getFrmDigitalE1PhysConfData(self, id):
    config = model.getObjectFromEditorId(id)
    digital = config.get_digitalE1()
    dchannel = digital.get_DChannelE1()
    audio = config.get_audio()
    data = {
      'e1NodeId': id,
      'id_textE1': config.get_ID(),
      'name_textE1': config.get_displayName(),
      #DigitalE1 values
      'signalingE1_selection': digital.get_E1Signaling(),
      'framingE1_selection': digital.get_framing(),
      'line_encE1_selection': digital.get_lineEncoding(),
      'terminationE1_selection': digital.get_E1Termination(),
      'clockingE1_selection': digital.get_clocking(),
      'loopback_modeE1_selection': digital.get_loopbackMode(),
      #Audio values
      'defaultPcmLawE1_selection': audio.get_defaultPcmLaw(),
      'voicePacketLengthInMsE1_text':audio.get_voicePacketLengthInMs(),
      'toneSetAttributesE1_select':audio.get_toneSetName(),
      'optimalTxBufferLevelE1_text':audio.get_optimalBchannelTxQueueLevel(),
      #EchoCancellation values
      'ecE1_type_selection':audio.get_EchoCancellation().get_type(),
      'ecE1_mode_selection':audio.get_EchoCancellation().get_mode(),
      'ecE1_comfortNoiseMode_selection':audio.get_EchoCancellation().get_comfortNoiseMode(),
      'ecE1_tailDisplacementInMs_text':audio.get_EchoCancellation().get_tailDisplacementInMs(),
      'ecE1_doubleTalkBehavior_selection':audio.get_EchoCancellation().get_doubleTalkBehavior(),
      #Acoustic values
      'aecE1_type_selection':audio.get_AcousticEchoCancellation().get_type(),
      'aecE1_mode_selection':audio.get_AcousticEchoCancellation().get_mode(),
      #LevelControl values
      'levelE1_type_selection':audio.get_LevelControl().get_type(),
      'levelE1_direction_selection':audio.get_LevelControl().get_direction(),
      'levelE1_mode_selection':audio.get_LevelControl().get_mode(),
      'levelE1_targetLevelIndBM0_text':audio.get_LevelControl().get_targetLevelIndBM0(),
      'levelE1_gainIndB_text':audio.get_LevelControl().get_gainIndB(),
      #BackgroundNoiseAttenuation values
      'bnaE1_type_selection':audio.get_BackgroundNoiseAttenuation().get_type(),
      'bnaE1_direction_selection':audio.get_BackgroundNoiseAttenuation().get_direction(),
      'bnaE1_mode_selection':audio.get_BackgroundNoiseAttenuation().get_mode(),
      'bnaE1_attenuationIndB_selection':audio.get_BackgroundNoiseAttenuation().get_attenuationIndB(),
      #DTMFRemoval values
      'dtmfE1_type_selection':audio.get_DTMFRemoval().get_type(),
      'dtmfE1_direction_selection':audio.get_DTMFRemoval().get_direction(),
      'dtmfE1_mode_selection':audio.get_DTMFRemoval().get_mode()
    }
    #DChannelE1 values
    if dchannel is None:
      data['disableDChannelE1'] = 'DISABLED'
    else:
      data['disableDChannelE1'] = 'ENABLED'
      data['channelE1_idx_text'] = dchannel.get_channelIndex()
      data['mtu_sizeE1_text'] = dchannel.get_mtuSize()
      data['record_to_pcap_fileE1'] = dchannel.get_RecordToPcapFile()
      data['pcap_files_folderE1'] = dchannel.get_PcapFilesFolder()

    log.debug("GetFrmDigitalE1PhysConfData data: %s" % data)
    return(str(extjs.FormActionLoadResponse(success=True, data=data)))

  ##########################################################################
  # Returns a list of all the values associated to the selected J1 config.
  # The fields to be exposed to the user are:
  #   - registered id (hidden field).
  #   - The display name of the configs
  #   - DigitalJ1 values
  #   - DChannelJ1 values
  #   - Audio values including EchoCancellation, AcousticEchoCancellation, etc.
  @expose() # Ext.form.Action.load - expects JSON
  def getFrmDigitalJ1PhysConfData(self, id):
    config = model.getObjectFromEditorId(id)
    digital = config.get_digitalJ1()
    dchannel = digital.get_DChannelJ1()
    audio = config.get_audio()
    data = {
      'j1NodeId': id,
      'id_textJ1': config.get_ID(),
      'name_textJ1': config.get_displayName(),
      #DigitalJ1 values
      'framingJ1_selection': digital.get_framing(),
      'line_encJ1_selection': digital.get_lineEncoding(),
      'lboJ1_selection': digital.get_LBO(),
      'clockingJ1_selection': digital.get_clocking(),
      'loopback_modeJ1_selection': digital.get_loopbackMode(),
      #Audio values
      'defaultPcmLawJ1_selection': audio.get_defaultPcmLaw(),
      'voicePacketLengthInMsJ1_text':audio.get_voicePacketLengthInMs(),
      'toneSetAttributesJ1_select':audio.get_toneSetName(),
      'optimalTxBufferLevelJ1_text':audio.get_optimalBchannelTxQueueLevel(),
      #EchoCancellation values
      'ecJ1_type_selection':audio.get_EchoCancellation().get_type(),
      'ecJ1_mode_selection':audio.get_EchoCancellation().get_mode(),
      'ecJ1_comfortNoiseMode_selection':audio.get_EchoCancellation().get_comfortNoiseMode(),
      'ecJ1_tailDisplacementInMs_text':audio.get_EchoCancellation().get_tailDisplacementInMs(),
      'ecJ1_doubleTalkBehavior_selection':audio.get_EchoCancellation().get_doubleTalkBehavior(),
      #Acoustic values
      'aecJ1_type_selection':audio.get_AcousticEchoCancellation().get_type(),
      'aecJ1_mode_selection':audio.get_AcousticEchoCancellation().get_mode(),
      #LevelControl values
      'levelJ1_type_selection':audio.get_LevelControl().get_type(),
      'levelJ1_direction_selection':audio.get_LevelControl().get_direction(),
      'levelJ1_mode_selection':audio.get_LevelControl().get_mode(),
      'levelJ1_targetLevelIndBM0_text':audio.get_LevelControl().get_targetLevelIndBM0(),
      'levelJ1_gainIndB_text':audio.get_LevelControl().get_gainIndB(),
      #BackgroundNoiseAttenuation values
      'bnaJ1_type_selection':audio.get_BackgroundNoiseAttenuation().get_type(),
      'bnaJ1_direction_selection':audio.get_BackgroundNoiseAttenuation().get_direction(),
      'bnaJ1_mode_selection':audio.get_BackgroundNoiseAttenuation().get_mode(),
      'bnaJ1_attenuationIndB_selection':audio.get_BackgroundNoiseAttenuation().get_attenuationIndB(),
      #DTMFRemoval values
      'dtmfJ1_type_selection':audio.get_DTMFRemoval().get_type(),
      'dtmfJ1_direction_selection':audio.get_DTMFRemoval().get_direction(),
      'dtmfJ1_mode_selection':audio.get_DTMFRemoval().get_mode()
    }
    #DChannelJ1 values
    if dchannel is None:
      data['disableDChannelJ1'] = 'DISABLED'
    else:
      data['disableDChannelJ1'] = 'ENABLED'
      data['channelJ1_idx_text'] = dchannel.get_channelIndex()
      data['mtu_sizeJ1_text'] = dchannel.get_mtuSize()
      data['record_to_pcap_fileJ1'] = dchannel.get_RecordToPcapFile()
      data['pcap_files_folderJ1'] = dchannel.get_PcapFilesFolder()

    return(str(extjs.FormActionLoadResponse(success=True, data=data)))
  
  ##########################################################################
  # Returns a list of all the values associated to the selected BRI config.
  # The fields to be exposed to the user are:
  #   - registered id (hidden field).
  #   - The display name of the configs
  #   - DigitalBRI values
  #   - DChannelBRI values
  #   - Audio values including EchoCancellation, AcousticEchoCancellation, etc.
  @expose() # Ext.form.Action.load - expects JSON
  def getFrmDigitalBRIPhysConfData(self, id):
    config = model.getObjectFromEditorId(id)
    digital = config.get_digitalBRI()
    dchannel = digital.get_DChannelBRI()
    audio = config.get_audio()
    data = {
      'briNodeId': id,
      'id_textBRI': config.get_ID(),
      'name_textBRI': config.get_displayName(),
      #DigitalBRI values
      'lboBRI_selection': digital.get_LBO(),
      'signalingBRI_selection': digital.get_BRISignaling(),
      'clockingBRI_selection': digital.get_clocking(),
      'loopback_modeBRI_selection': digital.get_loopbackMode(),
      #Audio values
      'defaultPcmLawBRI_selection': audio.get_defaultPcmLaw(),
      'voicePacketLengthInMsBRI_text':audio.get_voicePacketLengthInMs(),
      'toneSetAttributesBRI_select':audio.get_toneSetName(),
      'optimalTxBufferLevelBRI_text':audio.get_optimalBchannelTxQueueLevel(),
      #EchoCancellation values
      'ecBRI_type_selection':audio.get_EchoCancellation().get_type(),
      'ecBRI_mode_selection':audio.get_EchoCancellation().get_mode(),
      'ecBRI_comfortNoiseMode_selection':audio.get_EchoCancellation().get_comfortNoiseMode(),
      'ecBRI_tailDisplacementInMs_text':audio.get_EchoCancellation().get_tailDisplacementInMs(),
      'ecBRI_doubleTalkBehavior_selection':audio.get_EchoCancellation().get_doubleTalkBehavior(),
      #Acoustic values
      'aecBRI_type_selection':audio.get_AcousticEchoCancellation().get_type(),
      'aecBRI_mode_selection':audio.get_AcousticEchoCancellation().get_mode(),
      #LevelControl values
      'levelBRI_type_selection':audio.get_LevelControl().get_type(),
      'levelBRI_direction_selection':audio.get_LevelControl().get_direction(),
      'levelBRI_mode_selection':audio.get_LevelControl().get_mode(),
      'levelBRI_targetLevelIndBM0_text':audio.get_LevelControl().get_targetLevelIndBM0(),
      'levelBRI_gainIndB_text':audio.get_LevelControl().get_gainIndB(),
      #BackgroundNoiseAttenuation values
      'bnaBRI_type_selection':audio.get_BackgroundNoiseAttenuation().get_type(),
      'bnaBRI_direction_selection':audio.get_BackgroundNoiseAttenuation().get_direction(),
      'bnaBRI_mode_selection':audio.get_BackgroundNoiseAttenuation().get_mode(),
      'bnaBRI_attenuationIndB_selection':audio.get_BackgroundNoiseAttenuation().get_attenuationIndB(),
      #DTMFRemoval values
      'dtmfBRI_type_selection':audio.get_DTMFRemoval().get_type(),
      'dtmfBRI_direction_selection':audio.get_DTMFRemoval().get_direction(),
      'dtmfBRI_mode_selection':audio.get_DTMFRemoval().get_mode()
    }
    #DChannelBRI values
    if dchannel is None:
      data['disableDChannelBRI'] = 'DISABLED'
    else:
      data['disableDChannelBRI'] = 'ENABLED'
      data['channelBRI_idx_text'] = dchannel.get_channelIndex()
      data['mtu_sizeBRI_text'] = dchannel.get_mtuSize()
      data['record_to_pcap_fileBRI'] = dchannel.get_RecordToPcapFile()
      data['pcap_files_folderBRI'] = dchannel.get_PcapFilesFolder()

    log.debug("GetFrmDigitalBRIPhysConfData data: %s" % data)
    return(str(extjs.FormActionLoadResponse(success=True, data=data)))

  
  ##########################################################################
  # Returns a list of all the values associated to the selected Analog config.
  # The fields to be exposed to the user are:
  #   - registered id (hidden field).
  #   - The display name of the configs
  #   - Analog values
  #   - Audio values including EchoCancellation, AcousticEchoCancellation, etc.
  @expose() # Ext.form.Action.load - expects JSON
  def getFrmAnalogPhysConfData(self, id):
    config = model.getObjectFromEditorId(id)
    analog = config.get_analog()
    audio = config.get_audio()
    data = {
      'analogNodeId': id,
      'id_textAnalog': config.get_ID(),
      'name_textAnalog': config.get_displayName(),
      #Analog values
      'variantAnalog_selection':analog.get_variant(),
      'analogClocking_selection':analog.get_clocking(),
      'analogFxoRxGain_text':analog.get_fxoRxGainIndB(),
      'analogFxoTxGain_text':analog.get_fxoTxGainIndB(),
      'analogFxsRxGain_select':analog.get_fxsRxGainIndB(),
      'analogFxsTxGain_select':analog.get_fxsTxGainIndB(),
      'analogFxoBatteryDebounce_text':analog.get_fxoBatteryDebounceTimeInMs(),
      'analogFxoBatteryThreshold_text':analog.get_fxoBatteryThresholdInVolts(),
      'analogFxoFakePolarityEnabled_selection':analog.get_analogFxoFakePolarityEnabled(),
      'analogFxoFAkePolarityThreshold_text':analog.get_fxoPolarityThresholdInDbov(),
      #Audio values
      'defaultPcmLawAnalog_selection':audio.get_defaultPcmLaw(),
      'voicePacketLengthInMsAnalog_text':audio.get_voicePacketLengthInMs(),
      'toneSetAttributesAnalog_select':audio.get_toneSetName(),
      'optimalTxBufferLevelAnalog_text':audio.get_optimalBchannelTxQueueLevel(),
      #EchoCancellation values
      'ecAnalog_type_selection':audio.get_EchoCancellation().get_type(),
      'ecAnalog_mode_selection':audio.get_EchoCancellation().get_mode(),
      'ecAnalog_comfortNoiseMode_selection':audio.get_EchoCancellation().get_comfortNoiseMode(),
      'ecAnalog_tailDisplacementInMs_text':audio.get_EchoCancellation().get_tailDisplacementInMs(),
      'ecAnalog_doubleTalkBehavior_selection':audio.get_EchoCancellation().get_doubleTalkBehavior(),
      #Acoustic values
      'aecAnalog_type_selection':audio.get_AcousticEchoCancellation().get_type(),
      'aecAnalog_mode_selection':audio.get_AcousticEchoCancellation().get_mode(),
      #LevelControl values
      'levelAnalog_type_selection':audio.get_LevelControl().get_type(),
      'levelAnalog_direction_selection':audio.get_LevelControl().get_direction(),
      'levelAnalog_mode_selection':audio.get_LevelControl().get_mode(),
      'levelAnalog_targetLevelIndBM0_text':audio.get_LevelControl().get_targetLevelIndBM0(),
      'levelAnalog_gainIndB_text':audio.get_LevelControl().get_gainIndB(),
      #BackgroundNoiseAttenuation values
      'bnaAnalog_type_selection':audio.get_BackgroundNoiseAttenuation().get_type(),
      'bnaAnalog_direction_selection':audio.get_BackgroundNoiseAttenuation().get_direction(),
      'bnaAnalog_mode_selection':audio.get_BackgroundNoiseAttenuation().get_mode(),
      'bnaAnalog_attenuationIndB_selection':audio.get_BackgroundNoiseAttenuation().get_attenuationIndB(),
      #DTMFRemoval values
      'dtmfAnalog_type_selection':audio.get_DTMFRemoval().get_type(),
      'dtmfAnalog_direction_selection':audio.get_DTMFRemoval().get_direction(),
      'dtmfAnalog_mode_selection':audio.get_DTMFRemoval().get_mode()
    }
    return(str(extjs.FormActionLoadResponse(success=True, data=data)))

  ##########################################################################
  # Saves the new informations of the config.
  # All values are editable except the id.
  @expose()
  def savePhysicalConf (self, **data):
    log.info("Saving physical configuration: %s" % data)
    physConfData = simplejson.loads(data['physicalConfig'])

    config = model.getObjectFromEditorId(physConfData['config']['nodeId'])
    audio = config.get_audio()

    # Sections are: config, analog, audio, EchoCancellation, etc.
    # Values are the values of that section wich is a dictionary
    for section, values in physConfData.items():
      log.info("Section is: %s" % section)
      if section == 'config':
        object = config
      elif section == 'digitalT1':
        object = config.get_digitalT1()
      elif section == 'digitalE1':
        object = config.get_digitalE1()
      elif section == 'digitalJ1':
        object = config.get_digitalJ1()
      elif section == 'digitalBRI':
        object = config.get_digitalBRI()
      elif section == 'analog':
        object = config.get_analog()
      elif section == 'analogFxs':
        object = config.get_analog()
      elif section == 'analogFxo':
        object = config.get_analog()
        if values['analogFxoFakePolarityEnabled'] == "NO":
          object.set_fxoPolarityThresholdInDbov(None)
      elif section == 'audio':
        object = audio
      elif section == "AcousticEchoCancellation":
        object = audio.get_AcousticEchoCancellation()
        if values['mode'] == "NORMAL":
          audio.get_EchoCancellation().set_tailDisplacementInMs(0)
      elif section == 'LevelControl':
        object = audio.get_LevelControl()
        if values['mode'] == 'AUTOMATIC':
          audio.get_LevelControl().set_gainIndB(None)
        elif values['mode'] == 'MANUAL':
          audio.get_LevelControl().set_targetLevelIndBM0(None)
        elif values['mode'] == 'DISABLED':
          audio.get_LevelControl().set_gainIndB(None) 
          audio.get_LevelControl().set_targetLevelIndBM0(None)
      elif section == 'DChannelT1':
        if values['disableDChannel'] == 'DISABLED':
          config.get_digitalT1().set_DChannelT1(None)
          object = None
        else:
          dchannelT1 = config.get_digitalT1().get_DChannelT1()
          if dchannelT1 is None:
            dchannelT1 = model.DChannelT1.factory()
          values['channelIndex'] = 24
          values['PcapFilesFolder'] = '/logs/dchan-pcap-recording/'
          del values['disableDChannel']
          config.get_digitalT1().set_DChannelT1(dchannelT1)
          object = dchannelT1
      elif section == 'DChannelE1':
        if not values.has_key('disableDChannel') or \
        values['disableDChannel'] == 'DISABLED':
          config.get_digitalE1().set_DChannelE1(None)
          object = None
        else:
          dchannelE1 = config.get_digitalE1().get_DChannelE1()
          if dchannelE1 is None:
            dchannelE1 = model.DChannelE1.factory()
          values['channelIndex'] = 16
          values['PcapFilesFolder'] = '/logs/dchan-pcap-recording/'
          del values['disableDChannel']
          config.get_digitalE1().set_DChannelE1(dchannelE1)
          object = dchannelE1    
      elif section == 'DChannelJ1':
        if values['disableDChannel'] == 'DISABLED':
          config.get_digitalJ1().set_DChannelJ1(None)
          object = None
        else:
          dchannelJ1 = config.get_digitalJ1().get_DChannelJ1()
          if dchannelJ1 is None:
            dchannelJ1 = model.DChannelJ1.factory()
          values['channelIndex'] = 24
          values['PcapFilesFolder'] = '/logs/dchan-pcap-recording/'
          del values['disableDChannel']
          config.get_digitalJ1().set_DChannelJ1(dchannelJ1)
          object = dchannelJ1 
      elif section == 'DChannelBRI':
        if values['disableDChannel'] == 'DISABLED':
          config.get_digitalBRI().set_DChannelBRI(None)
          object = None
        else:
          dchannelBRI = config.get_digitalBRI().get_DChannelBRI()
          if dchannelBRI is None:
            dchannelBRI = model.DChannelBRI.factory()
          values['channelIndex'] = 3
          values['PcapFilesFolder'] = '/logs/dchan-pcap-recording/'
          del values['disableDChannel']
          config.get_digitalBRI().set_DChannelBRI(dchannelBRI)
          object = dchannelBRI
      else:
        getObjectMethod = getattr(audio, "get_%s" % section)
        object = getObjectMethod()

      if object:
        model.saveAttributes(object,values)
    # Write current config to file using default file location
    model.writeCurrentConfigToFile()

    return(str(extjs.FormActionSubmitResponse(success=True)))

  ##########################################################################
  # Returns a list of all the physical configs (E1, T1, Analog), available
  # in the system.
  # The fields to be exposed to the user are:
  #   - Real ID of the board in the config.
  #   - The display name of the configs
  @expose() # Ext.form.Action.load - expects JSON
  def getGridPhysicalConfigData(self, id):
    # Get the physicalConfig for digital boards
    data = []
    for physicalConfig in (model.currentConfig.get_physicalInterfaceConfigs().get_sangomaDigitalConfig() + \
      model.currentConfig.get_physicalInterfaceConfigs().get_sangomaAnalogConfig()):
      data.append({
        'id': physicalConfig.get_ID(),
        'displayName': physicalConfig.get_displayName(),
        'nodeId': model.getObjectId(physicalConfig)
      })
    data = sorted(data, key=itemgetter('id'))
    #data.sort()

    return(str(extjs.FormActionLoadResponse(success=True, data=data)))
#################################################
#                                               #
# ApplyTo PhysicalConfigs                       #
#                                               #
#################################################

  ##########################################################################
  # Returns a list of all the analog boards that are not assigned to a
  # physical configuration.
  # The fields to be exposed to the user are:
  #   - ID of the board.
  #   - The display name of the board
  @expose() # Ext.form.Action.load - expects JSON
  def getUnassignedBoards(self, id):
    log.info("Getting all the unassigned analog boards")
    isBoardAssigned = False
    data = []
    for board in model.currentConfig.get_boards().get_sangomaAnalogBoard():
      for config in model.currentConfig.get_physicalInterfaceConfigs().get_sangomaAnalogConfig():
        for interfaceRef in model.getInterfaceRefsForConfig(config):
          if interfaceRef.get_boardID() == board.get_ID():
            isBoardAssigned = True
            break
      if not(isBoardAssigned):
        data.append({
          'id':board.get_ID(),
          'displayName':board.get_displayName(),
          'configID':id
        })
      isBoardAssigned = False
    return(str(extjs.FormActionLoadResponse(success=True, data=data)))

  ##########################################################################
  # Returns a list of all the digital interfaces that are not assigned to a
  # physical configuration.
  # The fields to be exposed to the user are:
  #   - ID of the board.
  #   - ID of the interface
  #   - The display name of the board
  #   - The display name of the interface
  @expose() # Ext.form.Action.load - expects JSON
  def getUnassignedInterfaces(self, id):
    log.info("Getting all the unassigned digital interfaces")
    interfaceRefList = []
    data = []

    for config in model.currentConfig.get_physicalInterfaceConfigs().get_sangomaDigitalConfig():
      for interfaceRef in model.getInterfaceRefsForConfig(config):
        interfaceRefList.append((interfaceRef.get_boardID(),interfaceRef.get_interfaceID()))

    for board in (model.currentConfig.get_boards().get_sangomaDigitalBoard() + \
      model.currentConfig.get_boards().get_sangomaVirtualDigitalBoard()):
      for interface in board.get_interface():
        if (board.get_ID(),interface.get_ID()) not in interfaceRefList:
          data.append({
            'id':interface.get_ID(),
            'boardID':board.get_ID(),
            'displayName':board.get_displayName(),
            'ifName':interface.get_displayName(),
            'configID':id
          })
    return(str(extjs.FormActionLoadResponse(success=True, data=data)))

  ##########################################################################
  # Returns a list of all the analog boards that are assigned to the
  # physical configuration.
  # The fields to be exposed to the user are:
  #   - ID of the board.
  #   - The display name of the board
  @expose() # Ext.form.Action.load - expects JSON
  def getAssignedBoards(self, id):
    log.info("Getting all the analog boards assigned to the config %s" % id)
    data = []

    config = model.getConfigObject(model.currentConfig, int(id))
    for interfaceRef in model.getInterfaceRefsForConfig(config):
      board = model.getBoardObject(model.currentConfig, interfaceRef.get_boardID())
      data.append({
            'id':board.get_ID(),
            'displayName':board.get_displayName(),
            'configID':id
          })
    return(str(extjs.FormActionLoadResponse(success=True, data=data)))


  ##########################################################################
  # Returns a list of all the digital interfaces that are assigned to the
  # physical configuration.
  # The fields to be exposed to the user are:
  #   - ID of the board.
  #   - ID of the interface
  #   - The display name of the board
  #   - The display name of the interface
  @expose() # Ext.form.Action.load - expects JSON
  def getAssignedInterfaces(self, id):
    log.info("Getting all the digital interfaces assigned to the config %s" % id)
    data = []
    config = model.getConfigObject(model.currentConfig, int(id))
    for interfaceRef in model.getInterfaceRefsForConfig(config):
      board = model.getBoardObject(model.currentConfig, int(interfaceRef.get_boardID()))
      interface = model.getInterfaceObject(model.currentConfig, int(interfaceRef.get_boardID()),int(interfaceRef.get_interfaceID()))
      data.append({
        'id':interfaceRef.get_interfaceID(),
        'boardID':interfaceRef.get_boardID(),
        'displayName':board.get_displayName(),
        'ifName':interface.get_displayName(),
        'configID':id
      })
    return(str(extjs.FormActionLoadResponse(success=True, data=data)))

  ##########################################################################
  # Apply the boards to the given configuration.
  @expose() # Ext.form.Action.load - expects JSON
  def applyBoardsToConfig(self, **data):
    log.info("Apply the boards to the config: %s" % data)
    boardsData = simplejson.loads(data['gridData'])

    #Find the config object
    physConf = model.getConfigObject(model.currentConfig,int( simplejson.loads(data['configID']) ))
    physConf.set_applyTo(None) #Clear the applyTo section

    for boardInfo in boardsData:
      #Find the board object
      board = model.getBoardObject(model.currentConfig,int(boardInfo['id']))
      model.applyConfigToBoard(physConf,board)

    model.writeCurrentConfigToFile()
    return(str(extjs.FormActionResponse(success=True)))

  ##########################################################################
  # Apply the interfaces to the given configuration.
  @expose() # Ext.form.Action.load - expects JSON
  def applyInterfacesToConfig(self, **data):
    log.info("Apply the interfaces to the config: %s" % data)
    interfacesData = simplejson.loads(data['gridData'])

    #Find the config object
    physConf = model.getConfigObject(model.currentConfig, int( simplejson.loads(data['configID']) ))
    physConf.set_applyTo(None) #Clear the applyTo section

    for interfaceInfo in interfacesData:
      board = model.getBoardObject(model.currentConfig, int(interfaceInfo['boardID']))
      interface = model.getInterfaceObject(model.currentConfig, board.get_ID(), int(interfaceInfo['id']))
      model.applyConfigToInterface(physConf,board,interface)

    model.writeCurrentConfigToFile()
    return(str(extjs.FormActionResponse(success=True)))

