Ext.namespace("Paraxip.config","Paraxip.config.advanced");


var storeAdvancedConfigValues = new Ext.data.JsonStore({
  id      : 'storeAdvancedConfigValues',
  url     : Paraxip.baseUrl + 'config/getAdvancedProperties',
  root    : 'data',
  fields  : ['configured','type','name','value','description']
});

Paraxip.config.advanced.saveProperties = function(){
  var rec = Paraxip.config.advanced.advancedGrid.getStore().getModifiedRecords();
  var transmitArray = new Array();

  // Validate that we have modified records before sending request.
  if (rec.length < 1 ){
    return;
  }

  for (var i = 0, row; row = rec[i++];) {
    name  = row.data['name'];
    value = row.data['value'];
    var tmpArray = new Array();
    tmpArray.push(name,value);
    transmitArray.push(tmpArray);
  }

  Ext.Ajax.request({
    url: Paraxip.baseUrl + 'config/saveAdvancedProperties',
    method: 'GET',
    params: {
        properties:Ext.util.JSON.encode(transmitArray)
    },
    success: function(form, action){
      var responses = Ext.decode(form.responseText);
      Paraxip.config.advanced.advancedGrid.getStore().commitChanges();
      Paraxip.config.advanced.advancedGrid.getStore().reload();
      Paraxip.isConfigOutOfSynch(true);
    }
  });
};


// the check column is created using a custom plugin
var checkColumn = new Ext.ux.grid.CheckColumn({
  header: 'Configured?',
  dataIndex: 'configured',
  sortable: true,
  editable: false,
  width: 50
});

//checkColumn.setEditable(false);

var bfield = new Ext.form.Field({
  autoCreate: {tag: 'select', children: [
    {tag: 'option', value: 'true', html: 'true'},
    {tag: 'option', value: 'false', html: 'false'}
  ]},
 /* getValue : function(){
    console.log("getValue: %s ", this.value);
    return this.value;
  },*/
  listeners:{
    change:function(combo, newVal, oldVal) {
      console.log("change:  %s ",oldVal);
    },
    focus:function(combo) {
      console.log("focus: %s ",combo.value);
    },
    blur:function(combo) {
      console.log("blur:  %s ",combo.value);
    }
  }
});

var colModel = new Ext.grid.ColumnModel({
   columns: [
    {
      header: "Type",
      id: "type",
      hidden: true,
      sortable: true,
      dataIndex: 'type'
    },
    {
      header: "Name",
      id: "name",
      //width: 75,
      sortable: true,
      dataIndex: 'name'
    },
    {
      header: "Value",
      id: "value",
      //width: 100,
      sortable: true,
      dataIndex: 'value',
      editable: true
    },
    checkColumn,
    {
      header: "Description",
      id: "description",
      hidden: true,
      sortable: true,
      dataIndex: 'description'
    }
  ],
  editors: {
    'text': new Ext.grid.GridEditor(new Ext.form.TextField({})),
    'number': new Ext.grid.GridEditor(new Ext.form.NumberField({})),
    //'date': new Ext.grid.GridEditor(new Ext.form.DateField({})),
    'boolean':  new Ext.grid.GridEditor(bfield)
  },
  getCellEditor: function(colIndex, rowIndex) {
    var field = this.getDataIndex(colIndex);
    if (field == 'value') {
      var store = Paraxip.config.advanced.advancedGrid.getStore();
      var rec = store.getAt(rowIndex);
      var type = rec.get('type');
      if (type == "INT_TYPE" || type== "FLOAT_TYPE"){
        return this.editors['number'];
      }
      else if (type == "BOOL_TYPE"){
        return this.editors['boolean'];
      }
      else{
        return this.editors['text'];
      }

    }
    return Ext.grid.ColumnModel.prototype.getCellEditor.call(this, colIndex, rowIndex);
  }
});



Paraxip.config.advanced.advancedGrid = new ParaxipEditorGridPanelWithHelp({
  id: 'Paraxip.config.advanced.advancedGrid',
  title: 'Advanced Configuration',
  frame:true,
  cls: 'advancedGrid',
  autoScroll: true,
  stripeRows: true,
  enableHdMenu: false,
  enableColumnMove: false,
  hideHeaders: false,
  border: true,
  height: 360,
  helpUrl: "advanced-config/advanced_configuration_help.html",
  style: 'text-align:left;white-space: normal;',
  selModel    : new Ext.grid.RowSelectionModel(),
  store:  storeAdvancedConfigValues,
  plugins: checkColumn,
  cm: colModel,
  viewConfig :{forceFit:true,scrollOffset: 0, autoFill: true},
  buttons:[{
    id: 'Paraxip.config.advanced.saveButton',
    text: 'Apply',
    listeners: {
      'click':function(){
        Paraxip.config.advanced.saveProperties();
      }
    }
    },{
    id: 'Paraxip.config.advanced.refreshButton',
    text: 'Refresh',
    listeners: {
      'click':function(){
        Paraxip.config.advanced.advancedGrid.getStore().reload();
      }
    }
  }]
});

Ext.override(Ext.ToolTip, {
  onTargetOver : function(e){
    if(this.disabled || e.within(this.target.dom, true)){
      return;
    }
    var t = e.getTarget(this.delegate);
    if (t){
      this.triggerElement = t;
      this.clearTimer('hide');
      this.targetXY = e.getXY();
      this.delayShow();
    }
  },
  onMouseMove : function(e){
    var t = e.getTarget(this.delegate);
    if (t) {
      this.targetXY = e.getXY();
      if (t === this.triggerElement) {
        if(!this.hidden && this.trackMouse){
          this.setPagePosition(this.getTargetXY());
        }
      } else {
        this.hide();
        this.lastActive = new Date(0);
        this.onTargetOver(e);
      }
    } else if (!this.closable && this.isVisible()) {
      this.hide();
    }
  },
  hide: function(){
    this.clearTimer('dismiss');
    this.lastActive = new Date();
    delete this.triggerElement;
    Ext.ToolTip.superclass.hide.call(this);
  }
});

Paraxip.config.advanced.advancedGrid.loadGrid = function(){
  if( Paraxip.config.advanced.advancedGrid.rendered ) {
    Paraxip.config.advanced.advancedGrid.getStore().load();
  }
};

Paraxip.config.advanced.advancedGrid.on(
  'render', function(grid){
    // load grid content
    Paraxip.config.advanced.advancedGrid.loadGrid();
    var tpl = new Ext.Template(
      '<b>{name}: </b><br>',
      '{description}');

    // tooltip code
    var store = grid.getStore();  // Capture the Store.

    var view = grid.getView();    // Capture the GridView.

    Paraxip.status.logging.globalLoggerConfigPanel.tip = new Ext.ToolTip({
      target: view.mainBody,    // The overall target element.

      delegate: '.x-grid3-row', // Each grid row causes its own seperate
                                // show and hide.

      trackMouse: true,         // Moving within the row should not hide the tip.

      renderTo: document.body,  // Render immediately so that tip.body can be
                                // referenced prior to the first show.

      width: 500,

      listeners: {  // Change content dynamically depending on which
                    // element triggered the show.

        beforeshow: function updateTipBody(tip) {
          //tip.body.dom.innerHTML = "Over Row " + view.findRowIndex(tip.triggerElement);
          var row = view.findRowIndex(tip.triggerElement);
          var record = store.getAt(row);
          tip.body.update(tpl.apply(record.data));
        }
      }
    });
  });

Paraxip.config.advancedPanel = new Ext.form.FormPanel({
  id          : 'Paraxip.config.advancedPanel',
  layout      : 'fit',
  items       : [Paraxip.config.advanced.advancedGrid]
});
