Ext.namespace("Paraxip",
              "Paraxip.config",
              "Paraxip.config.callControl");
/*
 * Reload the treeview
 * @param reloadFromCurrentNode: true if we should reset focus to a specific node;
 *    false if not
 */
function reloadTree(reloadFromCurrentNode){
  var currentNode = treePanel.getSelectionModel().getSelectedNode(); //retrieve the current node
  treePanel.collapseAll();

  treePanel.root.reload(function(){
    if( reloadFromCurrentNode ) {
      goToNode( currentNode.id );
    } else {
      //Brings the emptyPanel view wich has nothing in it
      Ext.getCmp('content-panel').layout.setActiveItem(0);
    }
  });
};

/*
 * Bring the user to the node corresponding to the given id
 * @param id: id of the node to select
 */
function goToNode(id){
  var destinationNode = treePanel.getNodeById(id);
  var currentPath = destinationNode.getPath();
  var parentNode = destinationNode.parentNode;

  treePanel.expandPath(currentPath);
  //To render all the nodes, fire a click event on each parent
  //node until the root node is reached<
  if( !parentNode.isRoot ) {
    treePanel.fireEvent('click',parentNode); //simulate a click event to reload the form
  }
  treePanel.fireEvent('click',destinationNode); //simulate a click event to reload the form
  treePanel.getSelectionModel().select(destinationNode);
};


// This is an empty panel only used to manage the undefined click and the reloads of
// the tree. It could eventually contain a helper message to tell the client what it
// is possible to accomplish in the editor.
var emptyPanel = new Ext.Panel({
  id        : 'emtpy-panel',
  frame     : true,
  header    : false,
  layout    : 'absolute',
  autoLoad  : Paraxip.baseUrl + "static/editor_first_panel.html"
});
// This is the main content center region that will contain each example layout panel.
// It will be implemented as a CardLayout since it will contain multiple panels with
// only one being visible at any given time.

var contentPanel = new Ext.Panel( {
  id            : 'content-panel',
  region        : 'center', // this is what makes this panel into a region within the containing layout
  layout        : 'card',
  layoutConfig  : {
    deferredRender: true,
    layoutOnTabChange: true
  },
  defaults      : {
    hideMode: 'offsets'
  },
  margins       : '2 5 5 0',
  activeItem    : 0,
  border        : false
});

contentPanel.add(emptyPanel);
contentPanel.add(gridBoards);
contentPanel.add(panelDigitalBoard);
contentPanel.add(panelAnalogBoard);
contentPanel.add(panelVirtualDigitalBoard);
contentPanel.add(frmDigitalInterface);
contentPanel.add(frmAnalogInterface);
contentPanel.add(frmVirtualDigitalInterface);
contentPanel.add(gridPhysicalConfig);
contentPanel.add(frmDigitalT1PhysConf);
contentPanel.add(frmDigitalE1PhysConf);
contentPanel.add(frmDigitalJ1PhysConf);
contentPanel.add(frmDigitalBRIPhysConf);
contentPanel.add(frmAnalogPhysConf);
contentPanel.add(gridChannelGroup);
contentPanel.add(frmChannelGroup);
contentPanel.add(gridISDNConfiguration);
contentPanel.add(frmIsdnFas);
contentPanel.add(frmIsdnNfas);
contentPanel.add(frmIsdnBri);
contentPanel.add(gridCASConfiguration);
contentPanel.add(frmCasR2);
contentPanel.add(gridAnalogConfiguration);
contentPanel.add(frmAnalogFxo);
contentPanel.add(frmAnalogFxs);

Ext.Ajax.request({
  url: Paraxip.baseUrl + "getGCParameterValue",
  disableCaching: true,
  params: {"paramName":'netborder.oamWebServer.hideR2Options'},
  success: function(result){
    var response = Ext.util.JSON.decode(result.responseText);
    if( response.paramValue ) {
      contentPanel.remove(gridCASConfiguration);
      contentPanel.remove(frmCasR2);
    }
  }
});

// Go ahead and create the TreePanel now so that we can use it below
var treePanel = new Ext.tree.TreePanel({
  id            : 'tree-panel',
  region        : 'west',
  layout        : 'absolute',
  margins       : '2 2 0 2',
  // enableDD      : true,
  width         : 175,
  minSize       : 100,
  maxSize       : 250,
  // tree-specific configs:
  rootVisible   : false,
  lines         : false,
  singleExpand  : false,
  useArrows     : true,

  loader: new Ext.tree.TreeLoader({
    dataUrl   :Paraxip.baseUrl + "config/fetchTree",
    preloadChildren  : true
  }),

  root: new Ext.tree.AsyncTreeNode()
});

// Assign the changeLayout function to be called on tree node click.
// "n" is newly clicked-on node
treePanel.on('click', function(n){
  var sn = this.selModel.selNode || {}; // selNode is null on initial selection
  if(n.id != sn.id){

    // ignore clicks on currently selected node
    // if the form is defined we show it
    if(n.attributes.frm){
      // load the form and send the node id as a parameter

        //console.info('Component type :',Ext.getCmp(n.attributes.frm).getXType());
      if(n.attributes.frm == 'gridBoards') {
        Ext.getCmp('gridBoards').getStore().load({
          //url: 'getGridBoardsData',
          params: {'id':n.id}
        });
      }
      if(n.attributes.frm == 'panelAnalogBoard') {
        Ext.getCmp('frmAnalogBoard').load({
          url: Paraxip.baseUrl + "boards/getFrmAnalogBoardData",
          params: {'id':n.id}
        });
        Ext.getCmp('gridInterfacesAnalog').getStore().load({
           //url: 'getGridInterfacesData',
           params: {'id':n.id}
        });
      }
      if(n.attributes.frm == 'panelDigitalBoard') {
        Ext.getCmp('frmDigitalBoard').load({
          url: Paraxip.baseUrl + "boards/getFrmDigitalBoardData"
          ,params: {'id':n.id}
          ,success: function() {
            // load the available interfaces in the primary clock reference
            // field
            primaryClockReference.store.load({
              params: {
                "id":Ext.getCmp("Paraxip.config.boards.digitalBoardID").getValue()
              }
            });
          }
        });
        Ext.getCmp('gridInterfacesDigital').getStore().load({
           //url: 'getGridInterfacesData',
           params: {'id':n.id}
        });
      }
      if(n.attributes.frm == 'panelVirtualDigitalBoard') {
        Ext.getCmp('frmVirtualDigitalBoard').load({
          url: Paraxip.baseUrl + "boards/getFrmVirtualDigitalBoardData",
          params: {'id':n.id}
        });
      }
      if(n.attributes.frm == 'frmAnalogInterface') {
        Ext.getCmp('frmAnalogInterface').load({
          url: Paraxip.baseUrl + "boards/getFrmAnalogInterfaceData",
          params: {'id':n.id},
          success: function() {
            Paraxip.config.boards.showFxsAttributes.call();
          }
        });
      }
      if(n.attributes.frm == 'frmDigitalInterface') {
        Ext.getCmp('frmDigitalInterface').load({
          url: Paraxip.baseUrl + "boards/getFrmDigitalInterfaceData",
          params: {'id':n.id}
        });
      }
      if(n.attributes.frm == 'frmVirtualDigitalInterface') {
        Ext.getCmp('frmVirtualDigitalInterface').load({
          url: Paraxip.baseUrl + "boards/getFrmVirtualDigitalInterfaceData",
          params: {'id':n.id},
          success: function(){
            levelControlVisibleFields.call(Ext.getCmp('frmDigitalT1PhysConf'));
            if( Ext.getCmp('pcap_files_folderT1').rendered ) {
              isDChannelEnabled.call(Ext.getCmp('disableDChannelT1'));
            }
          }
        });
      }
      if(n.attributes.frm == 'frmDigitalT1PhysConf') {
        Ext.getCmp('frmDigitalT1PhysConf').load({
          url: Paraxip.baseUrl + "phys/getFrmDigitalT1PhysConfData",
          params: {'id':n.id},
          success: function(){
            levelControlVisibleFields.call(Ext.getCmp('frmDigitalT1PhysConf'));
            echoCancellationDisabled.call(Ext.getCmp('ecT1_mode_selection'));
            forceTailDisplacementValue.call(Ext.getCmp("aecT1_mode_selection"));
            if( Ext.getCmp('pcap_files_folderT1').rendered ) {
              isDChannelEnabled.call(Ext.getCmp('disableDChannelT1'));
            }
          }
        });
      }
      if(n.attributes.frm == 'frmDigitalE1PhysConf') {
        Ext.getCmp('frmDigitalE1PhysConf').load({
          url: Paraxip.baseUrl + "phys/getFrmDigitalE1PhysConfData",
          params: {'id':n.id},
          scope: Ext.getCmp('frmDigitalE1PhysConf'),
          success: function(){
            levelControlVisibleFields.call(Ext.getCmp('frmDigitalE1PhysConf'));
            echoCancellationDisabled.call(Ext.getCmp('ecE1_mode_selection'));
            forceTailDisplacementValue.call(Ext.getCmp("aecE1_mode_selection"));
            if( Ext.getCmp('pcap_files_folderE1').rendered ) {
              isDChannelEnabled.call(Ext.getCmp('disableDChannelE1'));
            }
          }
        });
      }
      if(n.attributes.frm == 'frmDigitalJ1PhysConf') {
        Ext.getCmp('frmDigitalJ1PhysConf').load({
          url: Paraxip.baseUrl + "phys/getFrmDigitalJ1PhysConfData",
          params: {'id':n.id},
          scope: Ext.getCmp('frmDigitalJ1PhysConf'),
          success: function(){
            levelControlVisibleFields.call(Ext.getCmp('frmDigitalJ1PhysConf'));
            echoCancellationDisabled.call(Ext.getCmp('ecJ1_mode_selection'));
            forceTailDisplacementValue.call(Ext.getCmp("aecJ1_mode_selection"));
            if( Ext.getCmp('pcap_files_folderJ1').rendered ) {
              isDChannelEnabled.call(Ext.getCmp('disableDChannelJ1'));
            }
          }
        });
      }
      
      if(n.attributes.frm == 'frmDigitalBRIPhysConf') {
        Ext.getCmp('frmDigitalBRIPhysConf').load({
          url: Paraxip.baseUrl + "phys/getFrmDigitalBRIPhysConfData",
          params: {'id':n.id},
          scope: Ext.getCmp('frmDigitalBRIPhysConf'),
          success: function(){
            levelControlVisibleFields.call(Ext.getCmp('frmDigitalBRIPhysConf'));
            echoCancellationDisabled.call(Ext.getCmp('ecBRI_mode_selection'));
            forceTailDisplacementValue.call(Ext.getCmp("aecBRI_mode_selection"));
            if( Ext.getCmp('pcap_files_folderBRI').rendered ) {
              isDChannelEnabled.call(Ext.getCmp('disableDChannelBRI'));
            }
          }
        });
      }
      if(n.attributes.frm == 'frmAnalogPhysConf') {
        Ext.getCmp('frmAnalogPhysConf').load({
          url: Paraxip.baseUrl + "phys/getFrmAnalogPhysConfData",
          params: {'id':n.id},
          scope: Ext.getCmp('frmAnalogPhysConf'),
          success: function(){
            levelControlVisibleFields.call(Ext.getCmp('frmAnalogPhysConf'));
            echoCancellationDisabled.call(Ext.getCmp('ecAnalog_mode_selection'));
            forceTailDisplacementValue.call(Ext.getCmp("aecAnalog_mode_selection"));
            isFxoFakePolarityEnabled.call(Ext.getCmp('analogFxoFakePolarityEnabled_selection'));
          }
        });
      }
      if(n.attributes.frm == 'gridPhysicalConfig') {
        Ext.getCmp('gridPhysicalConfig').getStore().load({
          params: {'id':n.id}
        });
      }
      if(n.attributes.frm == 'gridChannelGroup') {
        Ext.getCmp('gridChannelGroup').getStore().load({
          params: {'id':n.id}
        });
      }
      if(n.attributes.frm == 'frmChannelGroup') {
        Ext.getCmp('frmChannelGroup').load({
          url: Paraxip.baseUrl + "call/getFrmChannelGroupData",
          params: {'id':n.id},
          success: showOutboundScheme
        });
      }
      if(n.attributes.frm == 'gridISDNConfiguration') {
        Ext.getCmp('gridISDNConfiguration').getStore().load({
          params: {'id':n.id}
        });
        //load grid in case it hasn't been loaded yet.
        Ext.getCmp('gridBoards').getStore().load({
          //url: 'getGridBoardsData',
          params: {'id':n.id}
        });
      }
      if(n.attributes.frm == 'gridCASConfiguration') {
        Ext.getCmp('gridCASConfiguration').getStore().load({
          params: {'id':n.id}
        });
        //load grid in case it hasn't been loaded yet.
        Ext.getCmp('gridBoards').getStore().load({
          //url: 'getGridBoardsData',
          params: {'id':n.id}
        });
      }
      if(n.attributes.frm == 'frmIsdnFas') {
        Ext.getCmp('frmIsdnFas').load({
          url: Paraxip.baseUrl + "call/getFrmIsdnFasData",
          params: {'id':n.id},
          success: function(){
            isWaitFacilityDelayEnable.call(Ext.getCmp('callerNameLocationFas_select'));
            updateProgressTonesIndicator_combobox.call(Ext.getCmp('switchVariantFas_select'));

          }
        });
      }
      if(n.attributes.frm == 'frmIsdnNfas') {
        Ext.getCmp('frmIsdnNfas').load({
          url: Paraxip.baseUrl + "call/getFrmIsdnNfasData",
          params: {'id':n.id},
          success: function(){
            isWaitFacilityDelayEnable.call(Ext.getCmp('callerNameLocationNfas_select'));
            updateProgressTonesIndicator_combobox.call(Ext.getCmp('switchVariantNfas_select'));
          }
        });
      }
      if(n.attributes.frm == 'frmIsdnBri') {
        Ext.getCmp('frmIsdnBri').load({
          url: Paraxip.baseUrl + "call/getFrmIsdnBriData",
          params: {'id':n.id},
          success: function(){
            isWaitFacilityDelayEnable.call(Ext.getCmp('callerNameLocationBri_select'));
            updateProgressTonesIndicator_combobox.call(Ext.getCmp('switchVariantBri_select'));
            briTopologyVisibleFields.call(Ext.getCmp('topologyBri_select'));
            isOverlapDigitsReceptionEnabled.call(Ext.getCmp('overlapDigitsReceptionBri_select'));
          }
        });
      }
      if(n.attributes.frm == 'frmCasR2') {
        Ext.getCmp('frmCasR2').load({
          url: Paraxip.baseUrl + "call/getFrmCasR2Data",
          params: {'id':n.id},
          success: function(){
            //isWaitFacilityDelayEnable.call(Ext.getCmp('callerNameLocationFas_select'));
            //updateProgressTonesIndicator_combobox.call(Ext.getCmp('switchVariantFas_select'));
          }
        });
      }
      if(n.attributes.frm == 'gridAnalogConfiguration') {
        Ext.getCmp('gridAnalogConfiguration').getStore().load({
          params: {'id':n.id}
        });
        //load grid in case it hasn't been loaded yet.
        Ext.getCmp('gridBoards').getStore().load({
          //url: 'getGridBoardsData',
          params: {'id':n.id}
        });
      }
      if(n.attributes.frm == 'frmAnalogFxo') {
        Ext.getCmp('frmAnalogFxo').load({
          url: Paraxip.baseUrl + "call/getFrmAnalogFxoData",
          params: {'id':n.id},
          success: function(){
            isNumberofRingsVisible.call();
            isRingCadenceDelayVisible.call();
            enableConnectOutBoundSideOptions.call();
            isDialingOutActive.call();
          }
        });
      }
      if(n.attributes.frm == 'frmAnalogFxs') {
        Ext.getCmp('frmAnalogFxs').load({
          url: Paraxip.baseUrl + "call/getFrmAnalogFxsData",
          params: {'id':n.id},
          success: function(){
            isHotlineDialingEnable.call();
          }
        });
      }

      // show form
      Ext.getCmp('content-panel').layout.setActiveItem(n.attributes.frm);
    }
    else{
      Ext.getCmp('content-panel').layout.setActiveItem(0);
    }
  }
});


// this fixes the problem of the static load from an ajax source
// see http://extjs.com/forum/showthread.php?t=34902

Ext.tree.TreeLoader.override({
  processResponse : function(response, node, callback){
      var json = response.responseText;
      try {
          var o = eval('('+json+')');
          node.beginUpdate();
          for(var i = 0, len = o.length; i < len; i++){
              var n = this.createNode(o[i]);
              if(n){
                  node.appendChild(n);
                  if(this.preloadChildren){
                      this.doPreload(n);
                  }
              }
          }
          node.endUpdate();
          if(typeof callback == 'function'){
              callback(this, node);
          }
      }catch(e){
          this.handleFailure(response);
      }
  }
});


// Finally, build the main layout once all the pieces are ready.  This is also a good
// example of putting together a full-screen BorderLayout within a Panel.
var editorMainPanel = new Ext.Panel({
  id        : "Paraxip.config.editorMainPanel",
  layout    : 'border',
  title     : 'PSTN Config',
  hideMode  : 'offsets',
  defaults  : {
    autoScroll: true,
    style: 'text-align:left',
    hideMode: 'offsets'
  },
  items     : [treePanel, contentPanel]
});




