//-----------------------------------------------
// Add/Delete Analog configuration methods
//-----------------------------------------------
Ext.namespace("Paraxip.config.callControl",
              "Paraxip.config.callControl.fxs");

var addAnalogConfHandler = function(){
  addAnalogConfWindow.show();
};

function addAnalogConfig(){
  var form = addAnalogConfigForm.getForm();
  var el  = form.getEl();

  el.dom.action = Paraxip.baseUrl + "call/addAnalogConfig";
  el.dom.method = 'POST';
  var analogConfig = Ext.util.JSON.encode(form.getValues());
  form.submit({
    params: {'analogConfig':analogConfig},
    success: function(){
      Paraxip.isConfigOutOfSynch(true);
      addAnalogConfWindow.hide();
      gridAnalogConfiguration.store.reload();
      reloadTree(true);
    },
    failure: Paraxip.failureMessage
  });
};

var deleteAnalogConfHandler = function(){
  var selectedRecords =
    gridAnalogConfiguration.getSelectionModel().getSelections();

  if (selectedRecords !== []) {
    var analogConfigs = new Array();
    for (var i = 0; i < selectedRecords.length; i++) {
      analogConfigs.push(selectedRecords[i].data);
    }
    analogConfigs = Ext.util.JSON.encode(analogConfigs);
    Ext.Ajax.request({
      url: Paraxip.baseUrl + "call/deleteAnalogConfig",
      method: 'POST',
      disableCaching: false,
      params: {"analogConfigs":analogConfigs},
      success: function ()
      {
        Paraxip.isConfigOutOfSynch(true);
        for (var index = 0; index < selectedRecords.length; index++) {
          gridAnalogConfiguration.store.remove(selectedRecords[index]);
        }
        reloadTree(true);
      },
      failure: Paraxip.failureMessage
    });
  }
};

var analogConfigChoice = new Ext.form.ComboBox({
  id              : 'analogConfigChoice'
  ,fieldLabel     : "Analog Config Type"
  ,store          : ['FXO','FXS']
  ,forceSelection : true
  ,triggerAction  : 'all'
  ,mode           : 'local'
  ,typeAhead      : true
  ,value          : 'FXO'
  ,selectOnFocus  : true
  ,width          : 100
});

var addAnalogConfigForm = new Ext.FormPanel({
  id          : "addAnalogConfigForm"
  ,region     : 'center'
  ,labelWidth : 175
  ,labelPad   : 10
  ,bodyStyle  : 'padding:10px;'
  ,frame      : false
  ,header     : false
  ,footer     : false
  ,buttonAlign: "right"
  ,buttons    : [{
    id: "Paraxip.config.callControl.addConfigButton"
    ,text: "Add"
    ,listeners: {'click':addAnalogConfig}
  },{
    id: "Paraxip.config.callControl.cancelAddConfigButton"
    ,text: "Cancel"
    ,listeners: {'click':function(){addAnalogConfWindow.hide();}}
  }]
});

addAnalogConfigForm.add(analogConfigChoice);

var addAnalogConfWindow = new Ext.Window({
  id          : "addAnalogWindow"
  ,layout     : 'border'
  ,width      : 505
  ,height     : 150
  ,closeAction: 'hide'
  ,plain      : true
});

addAnalogConfWindow.add(addAnalogConfigForm);

var gridAnalogConfiguration = new Ext.grid.GridPanel({
  title       : 'Analog Configuration Grid'
  ,id         : 'gridAnalogConfiguration'
  ,autoScroll : true
  ,layout     : 'absolute'
  ,store      : new Ext.data.JsonStore({
    id      : 'storeAnalogConfiguration'
    ,url    : Paraxip.baseUrl + "call/getGridAnalogConfigurationData"
    ,root   : 'data'
    ,fields : [
      {name: 'id'},
      {name: 'displayName'},
      {name: 'nodeId'} //not displayed
    ]
  })
  ,columns    : [
    {
      header    : "ID"
      ,width    : 35
      ,sortable : true
      ,dataIndex: 'id'
      ,css      :'padding:5px;'
    },
    {
      header    : "Name"
      ,width    : 200
      ,sortable : true
      ,dataIndex: 'displayName'
      ,css      :'padding:5px;'
    }
  ]
  ,viewConfig:{forceFit:true}
});

gridAnalogConfiguration.addButton("Add Analog Config"
                                  ,addAnalogConfHandler
                                  ,gridAnalogConfiguration);
gridAnalogConfiguration.addButton("Remove Analog Config(s)"
                                  ,deleteAnalogConfHandler
                                  ,gridAnalogConfiguration);

gridAnalogConfiguration.on('rowdblclick', function(grid, rowIndex){
  var record = grid.store.getAt(rowIndex);
  goToNode(record.get('nodeId'));
});

function isNumberofRingsVisible(){
  var notifyOutCond = notifyOutbound_condition_select;
  var numberOfRings = numOfRings_text;
  if( notifyOutCond.getValue() == "AFTER-N-RING") {
    numberOfRings.enable();
  } else {
    numberOfRings.disable();
    numberOfRings.clearInvalid();
  }
};

function enableConnectOutBoundSideOptions(){
  var connectOutCond = connectOutBoundSide_condition_select;
  var disconnectTimeout = disconnectTimeoutMs_text;
  var connectOnNoRingTimeout = connectOnNoRemoteRingTimeoutMs_text;
  if( connectOutCond.getValue() == "ON-REMOTE-RING-STOPPED" ) {
    disconnectTimeout.enable();
    connectOnNoRingTimeout.enable();
  } else if( connectOutCond.getValue() == "ON-BATTERY-POLARITY-INVERSION" ) {
    disconnectTimeout.enable();
    connectOnNoRingTimeout.disable();
    connectOnNoRingTimeout.clearInvalid();
  } else {
    disconnectTimeout.disable();
    disconnectTimeout.clearInvalid();
    connectOnNoRingTimeout.disable();
    connectOnNoRingTimeout.clearInvalid();
  }
};

function isRingCadenceDelayVisible(){
  var ringCadenceOn = ringCadenceDetector_active_select;
  var ringCadenceDelay = ringCadenceDetector_OffDelayMs_text;

  if( ringCadenceOn.getValue() == "YES" ) {
    ringCadenceDelay.enable();
  } else {
    ringCadenceDelay.disable();
  }
};

function isDialingOutActive(){
  var dialingOutActive = dial_active_select;

  var waitForDialTone = dial_waitForDialTone_select;
  var preDialDelayInMs = preDialDelayInMs_text;
  var interDigitDelayInMs = interDigitDelayInMs_text;
  var dTMFDigitDurationInMs = dTMFDigitDurationInMs_text;

  var disconnectTimeoutMs = disconnectTimeoutMs_text;

  var outboundConnect = outboundConnectPanel;

  var intDialSpecCharacterSubs = intDialSpecCharacterSubs_text;
  var intDialSpecCharacter = intDialSpecCharacter_text;

  if( dialingOutActive.getValue() == "YES" ) {
    waitForDialTone.enable();
    preDialDelayInMs.enable();
    interDigitDelayInMs.enable();
    dTMFDigitDurationInMs.enable();
    intDialSpecCharacterSubs.enable();
    intDialSpecCharacter.enable();

    //Outbound Call - Connect Info params
    outboundConnect.enable();

    if( disconnectTimeoutMs.getValue() == "") {
      disconnectTimeoutMs.setValue( 40000 );
    }

  } else {
    waitForDialTone.disable();
    preDialDelayInMs.disable();
    preDialDelayInMs.clearInvalid();
    interDigitDelayInMs.disable();
    interDigitDelayInMs.clearInvalid();
    dTMFDigitDurationInMs.disable();
    dTMFDigitDurationInMs.clearInvalid();
    intDialSpecCharacterSubs.disable();
    intDialSpecCharacterSubs.clearInvalid();
    intDialSpecCharacter.disable();
    intDialSpecCharacter.clearInvalid();

    //Outbound Call - Connect Info params
    outboundConnect.disable();

  }
};

function isHotlineDialingEnable(){
  if(Paraxip.config.callControl.fxs.hotlineDialingActive.getValue() == "YES") {
    Paraxip.config.callControl.fxs.hotlineDelayBeforeDialing.enable();
    Paraxip.config.callControl.fxs.hotlineDialingNumber.enable();
  } else {
    Paraxip.config.callControl.fxs.hotlineDelayBeforeDialing.disable();
    Paraxip.config.callControl.fxs.hotlineDelayBeforeDialing.clearInvalid();
    Paraxip.config.callControl.fxs.hotlineDialingNumber.disable();
    Paraxip.config.callControl.fxs.hotlineDialingNumber.clearInvalid();
  }
}

//------------------------------------------------
// Analog FXO config form
// -----------------------------------------------
function saveFxoConfig(e){
  var owner = e.ownerCt;
  var form = owner.getForm();
  var el = form.getEl();

  el.dom.action = Paraxip.baseUrl + "call/saveFxoConfig";
  el.dom.method = 'POST';

  var fxoConfig = owner.getItemsValueAsJson();

  form.submit({
    params:{
      'frmAnalogFxo': fxoConfig
    },
    success: function(){
      Paraxip.isConfigOutOfSynch(true);
      reloadTree(true);
    },
    failure: Paraxip.failureMessage
  });
};

var frmAnalogFxo = new Ext.FormPanel({
  title         : "Analog FXO config"
  ,frame        : true
  ,header       : false
  ,id           : "frmAnalogFxo"
  ,autoScroll   : true
  ,layout       : 'accordion'
  ,layoutConfig : {
    activeItem  : 0
  }
  ,defaults     : {
    layout      : 'form'
    ,style      : 'text-align:left;'
    ,labelWidth : 150
    ,bodyStyle  : 'padding:5px;'
    ,width      : 500
    ,autoHeight : true
    ,allowBlank : false
    ,fill       : true
  }
  ,buttonAlign  : "right"
  ,buttons      : [{
    id        : "Paraxip.config.callControl.assignedInterfaceButton"
    ,text     : 'Assigned Interface(s)'
    ,listeners: {
      'click':function(){
        fxoConfigID = idFxo_text.getValue();
        analogFxoInterfaceGrid.getStore().load({
          url: Paraxip.baseUrl + "call/getAssignedFxoInterfaces"
          ,params: {'id':fxoConfigID}
        });
        storeAnalogFxoBoards.load({
          url: Paraxip.baseUrl + "call/getAnalogFxoBoards"
          ,params: {'id':fxoConfigID}
        });
        storeAnalogFxoChannelGroup.load({
          url: Paraxip.baseUrl + "call/getChannelGroups"
          ,params: {'id':fxoConfigID}
        });
        assignAnalogFxoInterfaceWindow.show();
      }
    }
  },{
    id        : "Paraxip.config.callControl.saveFxoConfigButton"
    ,text     : 'Save'
    ,listeners: {'click':saveFxoConfig}
  },{
    id        : "Paraxip.config.callControl.cancelSaveConfigButton"
    ,text     : 'Cancel'
    ,listeners: {'click':callControlCancel}
  }]
});

var fxoConfigPanel = new ParaxipPanelWithHelp({
  id    : 'fxoConfigPanel'
  ,name : 'fxoConfig'
  ,title: 'Configuration Info'
  ,items: [{
    xtype     : 'textfield'
    ,name     : 'nodeId'
    ,hidden   : true
    ,hideLabel: true
  }]
  ,helpText: "analogFxoConfiganalogFxoConfig"
});

var inboundCallHandlingPanel = new ParaxipPanelWithHelp({
  id        : 'inboundCallHandlingPanel'
  ,name     : 'inboundCall'
  ,title    : 'Inbound Call - Handling Info'
  ,helpText : "inboundCallParametersinboundCallHandling"
});

var outboundCallDialingPanel = new ParaxipPanelWithHelp({
  id        : 'outboundCallDialingPanel'
  ,name     : 'outboundCallDialing'
  ,title    : 'Outbound Call - Dialing Info'
  ,helpText : "dialParametersdial"
});

var outboundConnectPanel = new ParaxipPanelWithHelp({
  id        : 'outboundConnectPanel'
  ,name     : 'outboundCallConnect'
  ,title    : 'Outbound Call - Connect Info'
  ,helpText : "outboundCallParametersoutboundCallHandling"
});

var disconnectSupervisionPanel = new ParaxipPanelWithHelp({
  id          : 'disconnectSupervisionPanel'
  ,name       : 'disconnectSupervision'
  ,title      : 'Disconnect Supervision Info'
  ,helpText   : "disconnectSupervisionParametersdisconnectSupervision"
});

var idFxo_text = createIdText({
  id        : 'idFxo_text'
  ,helpText : 'analogFxoConfiganalogFxoConfigID'
});

var nameFxo_text = createNameText({
  id        : 'nameFxo_text'
  ,helpText : 'analogFxoConfiganalogFxoConfigdisplayName'
});

var minimumOnHookDurationMs = createMinimumOnHookDurationMs({
  id        : 'minimumOnHookDurationMs'
  ,helpText : 'analogFxoConfiganalogFxoConfigminimumOnHookDurationMs'
});

var numOfRings_text = new Ext.form.NumberField({
  id            : 'numOfRings_text'
  ,fieldLabel   : "Number Of Rings"
  ,name         : "numberOfRings"
  ,helpText     : "notifyOutboundSideParametersnotifyOutboundSidenumberOfRings"
  ,allowNegative: false
  ,allowBlank   : false
  ,disabled     : true
  ,width        : 217
});

var notifyOutbound_condition_select = new Ext.form.ComboBox({
  id              : 'notifyOutbound_condition_select'
  ,name           : 'notifyOutbound_condition'
  ,helpText       : "notifyOutboundSideParametersnotifyOutboundSide"
  ,fieldLabel     : "Notify Outbound Side Condition"
  ,store          : ['ON-FIRST-RING','ON-CALLER-ID','AFTER-N-RING']
  ,forceSelection : true
  ,triggerAction  : 'all'
  ,mode           : 'local'
  ,typeAhead      : true
  ,selectOnFocus  : true
  ,width          : 200
  ,minListWidth   : 200
  ,listeners      : {
    'select':isNumberofRingsVisible
  }
});

var notifyOutbound_delay_text = new Ext.form.NumberField({
  id              : 'notifyOutbound_delay_text'
  ,fieldLabel     : "Notify Outbound Side Delay (ms)"
  ,name           : "notifyOutbound_delay"
  ,helpText       : "notifyOutboundSideParametersnotifyOutboundSidedelayInMs"
  ,allowNegative  : false
  ,allowBlank     : false
  ,width          : 217
});

var notifyOutboundConnect_condition_select = new Ext.form.ComboBox({
  id              : 'notifyOutboundConnect_condition_select'
  ,name           : 'connect_condition'
  ,fieldLabel     : "Connect Inbound Side Condition"
  ,helpText       : "connectParametersInboundconnect"
  ,store          : ['ON-OUTBOUNDSIDE-NOTIFIED','ON-OUTBOUNDSIDE-CONNECTED',
                     'ON-OUTBOUNDSIDE-MEDIA-AVAILABLE']
  ,forceSelection : true
  ,triggerAction  : 'all'
  ,mode           : 'local'
  ,typeAhead      : true
  ,selectOnFocus  : true
  ,width          : 200
  ,minListWidth   : 200
});

var callerID_type_select = new Ext.form.ComboBox({
  id              : 'callerID_type_selection'
  ,name           : 'callerID_type'
  ,fieldLabel     : "CallerID Type"
  ,helpText       : "callerIDParameterscallerID"
  ,store          : ['BT-Bellcore-FSK-SDMF-MDMF']
  ,forceSelection : true
  ,triggerAction  : 'all'
  ,mode           : 'local'
  ,typeAhead      : true
  ,selectOnFocus  : true
  ,width          : 200
  ,minListWidth   : 200
});

var dial_active_select = new Ext.form.ComboBox({
  id              : 'dial_active_selection'
  ,name           : 'dial_active'
  ,fieldLabel     : "Enable Dialing Out"
  ,helpText       : "dialParametersdialactive"
  ,store          : ['YES','NO']
  ,forceSelection : true
  ,triggerAction  : 'all'
  ,mode           : 'local'
  ,typeAhead      : true
  ,selectOnFocus  : true
  ,width          : 235
  ,minListWidth   : 235
  ,listeners      : {
    'select':isDialingOutActive
  }
});

var dial_waitForDialTone_select = new Ext.form.ComboBox({
  id              : 'dial_waitForDialTone_selection'
  ,name           : 'dial_waitForDialTone'
  ,fieldLabel     : "Wait For Dial Tone"
  ,helpText       : "dialParametersdialwaitForDialTone"
  ,store          : ['YES','NO']
  ,forceSelection : true
  ,triggerAction  : 'all'
  ,mode           : 'local'
  ,typeAhead      : true
  ,selectOnFocus  : true
  ,width          : 235
  ,minListWidth   : 235
});


var intDialSpecCharacter_text = new Ext.form.TextField({
  id              : 'intDialSpecCharacter_text'
  ,fieldLabel     : "International Dialing Special Character"
  ,name           : "intDialSpecCharacter"
  ,helpText       : "dialParametersdialInternationalDialingSpecialCharacter"
  ,allowBlank     : false
  ,maskRe         : /[^0-9]/
  ,autoCreate: {tag: 'input', type: 'text', autocomplete: 'off', maxlength: '1'}
  ,width          : 252
});


var intDialSpecCharacterSubs_text = new Ext.form.TextField({
  id              : 'intDialSpecCharacterSubs_text'
  ,fieldLabel     : "International Dialing Special Character Substitute"
  ,name           : "intDialSpecCharacterSubs"
  ,helpText       : "dialParametersdialInternationalDialingSpecialCharacterSubstitute"
  ,allowBlank     : false
  ,maskRe         : /[0-9]/
  ,width          : 252
});


var preDialDelayInMs_text = new Ext.form.NumberField({
  id              : 'preDialDelayInMs_text'
  ,fieldLabel     : "Pre Dial Delay (ms)"
  ,name           : "preDialDelayInMs"
  ,helpText       : "dialParametersdialpreDialDelayInMs"
  ,allowNegative  : false
  ,allowBlank     : false
  ,width          : 252
});

var interDigitDelayInMs_text = new Ext.form.NumberField({
  id              : 'interDigitDelayInMs_text'
  ,fieldLabel     : "InterDigit Delay (ms)"
  ,name           : "interDigitDelayInMs"
  ,helpText       : "dialParametersdialinterDigitDelayInMs"
  ,allowNegative  : false
  ,allowBlank     : false
  ,width          : 252
});

var dTMFDigitDurationInMs_text = new Ext.form.NumberField({
  id              : 'dTMFDigitDurationInMs_text'
  ,fieldLabel     : "DTMF Digit Duration (ms)"
  ,name           : "dTMFDigitDurationInMs"
  ,helpText       : "dialParametersdialDTMFDigitDurationInMs"
  ,allowNegative  : false
  ,allowBlank     : false
  ,width          : 252
});


var disconnectTimeoutMs_text = new Ext.form.NumberField({
  id              : 'disconnectTimeoutMs_text'
  ,fieldLabel     : "Disconnect Timeout (ms)"
  ,name           : "disconnectTimeoutMs"
  ,helpText       : "connectSupervisionParametersconnectSupervisionconnect" +
                    "OutboundSidedisconnectTimeoutMs"
  ,value          : 40000
  ,allowNegative  : false
  ,allowBlank     : false
  ,disabled       : true
  ,width          : 217
});

var connectOnNoRemoteRingTimeoutMs_text = new Ext.form.NumberField({
  id              : 'connectOnNoRemoteRingTimeoutMs_text'
  ,fieldLabel     : "Connect On No Remote Ring Timeout(ms)"
  ,name           : "connectOnNoRemoteRingTimeoutMs"
  ,helpText       : "connectSupervisionParametersconnectSupervisionconnect" +
                    "OutboundSideconnectOnNoRemoteRingTimeoutMs"
  ,allowNegative  : false
  ,allowBlank     : false
  ,disabled       : true
  ,width          : 217
});

var connectOutBoundSide_condition_select = new Ext.form.ComboBox({
  id              : 'connectOutBoundSide_condition_selection'
  ,name           : 'connectOutBoundSide_condition'
  ,fieldLabel     : "Connect Outbound Side Condition"
  ,helpText       : "connectSupervisionParametersconnectSupervisionconnect" +
                    "OutboundSidecondition"
  ,store          : ['AFTER-DIALING-DESTINATION-NUMBER'
                     ,'ON-REMOTE-RING-STOPPED'
                     /*,'ON-VOICE'*/ //ON-VOICE not supported for now
                     ,'ON-BATTERY-POLARITY-INVERSION']
  ,forceSelection : true
  ,triggerAction  : 'all'
  ,mode           : 'local'
  ,typeAhead      : true
  ,selectOnFocus  : true
  ,width          : 235
  ,minListWidth   : 235
  ,listeners      : {
    'select':enableConnectOutBoundSideOptions
  }
});

var connectParamOutbound_condition_select = new Ext.form.ComboBox({
  id              : 'connectParamOutbound_condition_selection',
  name            : 'connectParamOutbound_condition',
  fieldLabel      : "Connect Inbound Side Condition",
  helpText        : "connectParametersOutboundconnectcondition",
  store           : ['ON-OUTBOUNDSIDE-CONNECTED','ON-OUTBOUNDSIDE-NOTIFIED',
                     'ON-INBOUNDSIDE-NOTIFICATION'],
  forceSelection  : true,
  triggerAction   : 'all',
  mode            : 'local',
  typeAhead       : true,
  selectOnFocus   : true,
  width           : 215,
  minListWidth    : 215
});

var connectParamOutbound_delayInMs_text = new Ext.form.NumberField({
  id            : 'connectParamOutbound_delayInMs_text',
  fieldLabel    : "Connect Inbound Side Delay (ms)",
  name          : "connectParamOutbound_delayInMs",
  helpText      : "connectParametersOutboundconnectdelayInMs",
  allowNegative : false,
  allowBlank    : false,
  width         : 232
});

var ringCadenceDetector_active_select = new Ext.form.ComboBox({
  id              : 'ringCadenceDetector_active_selection',
  name            : 'ringCadenceDetector_active',
  fieldLabel      : "Ring Cadence Detector Active",
  helpText        : "disconnectSupervisionParametersdisconnectSupervisiondisconnectDetectorsringCadenceDetectoractive",
  store           : ['YES','NO'],
  forceSelection  : true,
  triggerAction   : 'all',
  mode            : 'local',
  typeAhead       : true,
  selectOnFocus   : true,
  width           : 145,
  minListWidth    : 145,
  listeners       : {
    'select':isRingCadenceDelayVisible
  }
});

var ringCadenceDetector_OffDelayMs_text = new Ext.form.NumberField({
  id            : 'ringCadenceDetector_OffDelayMs_text',
  fieldLabel    : "Ring Cadence Off Delay (ms)",
  name          : "ringCadenceDetector_OffDelayMs",
  helpText      : "disconnectSupervisionParametersdisconnectSupervisiondisconnectDetectorsringCadenceDetectorringCadenceOffDelayMs",
  allowNegative : false,
  allowBlank    : false,
  width         : 162
});

var disconnectDetectors_checkGroup = new Ext.form.CheckboxGroup({
    id:'disconnectDetectors_group',
    name: 'disconnectDetectors',
    xtype: 'checkboxgroup',
    fieldLabel: 'Disconnect call on',
    itemCls: 'x-check-group-alt',
    // Put all controls in a single column with width 100%
    columns: 1,
    helpText: "disconnectSupervisionParametersdisconnectSupervisiondisconnectDetectorsdetector",
    items: [
      {boxLabel: 'Battery Disconnect',
          name: 'batteryDisconnect',
          id: 'batteryDisconnectDetector_chk',
          checked: false},
      {boxLabel: 'Battery Polarity Inversion',
          name: 'batteryPolarityInversion',
          id: 'batteryPolarityInversionDetector_chk',
          checked: false},
      {boxLabel: 'Dial tone',
          name: 'dial',
          id: 'dialToneDetector_chk',
          checked: false},
      {boxLabel: 'Reorder tone',
          name: 'reorder',
          id: 'reorderToneDetector_chk',
          checked: false},
      {boxLabel: 'Busy tone',
          name: 'busy',
          id: 'busyToneDetector_chk',
          checked: false},
      {boxLabel: 'Sit tone',
          name: 'sit',
          id: 'sitToneDetector_chk',
          checked: false},
      {boxLabel: 'Disconnect tone',
          name: 'disconnect',
          id: 'disconnectToneDetector_chk',
          checked: false}
    ]
});

fxoConfigPanel.add(idFxo_text);
fxoConfigPanel.add(nameFxo_text);
fxoConfigPanel.add(minimumOnHookDurationMs);
frmAnalogFxo.add(fxoConfigPanel);

inboundCallHandlingPanel.add(notifyOutbound_condition_select);
inboundCallHandlingPanel.add(numOfRings_text);
inboundCallHandlingPanel.add(notifyOutbound_delay_text);
inboundCallHandlingPanel.add(notifyOutboundConnect_condition_select);
inboundCallHandlingPanel.add(callerID_type_select);
frmAnalogFxo.add(inboundCallHandlingPanel);

outboundCallDialingPanel.add(dial_active_select);
outboundCallDialingPanel.add(dial_waitForDialTone_select);
outboundCallDialingPanel.add(intDialSpecCharacter_text);
outboundCallDialingPanel.add(intDialSpecCharacterSubs_text);
outboundCallDialingPanel.add(preDialDelayInMs_text);
outboundCallDialingPanel.add(interDigitDelayInMs_text);
outboundCallDialingPanel.add(dTMFDigitDurationInMs_text);
frmAnalogFxo.add(outboundCallDialingPanel);

outboundConnectPanel.add(connectOutBoundSide_condition_select);
outboundConnectPanel.add(disconnectTimeoutMs_text);
outboundConnectPanel.add(connectOnNoRemoteRingTimeoutMs_text);
outboundConnectPanel.add(connectParamOutbound_condition_select);
outboundConnectPanel.add(connectParamOutbound_delayInMs_text);
frmAnalogFxo.add(outboundConnectPanel);

disconnectSupervisionPanel.add(ringCadenceDetector_active_select);
disconnectSupervisionPanel.add(ringCadenceDetector_OffDelayMs_text);
disconnectSupervisionPanel.add(disconnectDetectors_checkGroup);
frmAnalogFxo.add(disconnectSupervisionPanel);

Paraxip.config.callControl.isIDInStore = function(store,id){
  var isValid = false;
  store.each(function(record){
    if( record.data.id == id ){
      isValid = true;
      return;
    }
  });

  return isValid;
}

Paraxip.config.callControl.validateID = function(id){
  if( !Paraxip.config.callControl.isIDInStore(this.store, id) ) {
    return "This ID can not be found in the list."
  }
  return true;
}

function applyToAnalogConfig(e){
  var grid = e.ownerCt;
  var url;
  var params;
  var gridData;
  var records;
  var i;

  records = grid.getStore().getRange();
  gridData = new Array();
  for (i = 0; i < records.length; i++) {
    gridData.push(records[i].data);
  }
  gridData = Ext.util.JSON.encode(gridData);

  if( grid.getItemId() == "analogFxoInterfaceGrid") {
    url = Paraxip.baseUrl + "call/applyInterfacesToFxoConfig";
    params = {
      'gridData':gridData,
      'fxoConfigID':Ext.util.JSON.encode(fxoConfigID)
    };
  } else if( grid.getItemId() == "analogFxsInterfaceGrid") {
    url = Paraxip.baseUrl + "call/applyInterfacesToFxsConfig";
    params = {
      'gridData':gridData,
      'fxsConfigID':Ext.util.JSON.encode(fxsConfigID)
    };
  }

  Ext.Ajax.request({
    url: url,
    method: 'POST',
    disableCaching: false,
    params: params,
    success: function(result){
      var response = Ext.util.JSON.decode(result.responseText);
      if(response.success) {
        Paraxip.isConfigOutOfSynch(true);
        grid.ownerCt.hide();
      } else {
        Ext.Msg.alert(
          "Invalid field(s)",
          "At least one of the fields in the entries is invalid."
        );
      }
    },
    failure: Paraxip.failureMessage
  });

};

function addAnalogFxoInterface(e){
  var form = e.ownerCt;

  if( storeAnalogFxoBoards.data.length < 1 ) {
    Ext.Msg.alert(
      "No Analog boards in config","There is no analog boards in your " +
      "configuration. Please add a board in the board section if you want " +
      "to assign a board to this config.");
    return;
  }

  var firstBoardId = storeAnalogFxoBoards.getAt(0).get('id');
  storeAnalogFxoInterfaces.load({
    url: Paraxip.baseUrl + "call/getAnalogFxoInterfaces",
    disableCaching: false,
    params: {'id':firstBoardId},
    callback: function(result){
      if( result.length > 0 ) {
        var newRecord = new Ext.data.Record({
          'boardId':firstBoardId,
          'interfaceId':storeAnalogFxoInterfaces.getAt(0).get('id'),
          'channelGroupID':storeAnalogFxoChannelGroup.getAt(0).get('id')
        });
        analogFxoInterfaceGrid.getStore().add(newRecord);
        analogFxoInterfaceGrid.expand();
      }
    }

  });
};

function deleteAnalogInterface(e){
  var grid = e.ownerCt;

  var selectedRecords = grid.getSelectionModel().getSelections();

  if (selectedRecords !== []) {
    for( i = 0; i < selectedRecords.length; i++ ) {
      grid.getStore().remove(selectedRecords[i]);
    }
  }
};

var fxoConfigID = '';

var storeAnalogFxoBoards = new Ext.data.JsonStore({
  id: 'storeAnalogFxoBoards',
  url: Paraxip.baseUrl + "call/getAnalogFxoBoards",
  root: 'data',
  fields: ['id','displayName','nodeId']
});

var storeAnalogFxoInterfaces = new Ext.data.JsonStore({
  id: 'storeAnalogFxoInterfaces',
  url: Paraxip.baseUrl + "call/getAnalogFxoInterfaces",
  root: 'data',
  fields: ['id','displayName']
});

var storeAnalogFxoChannelGroup = new Ext.data.JsonStore({
  id: 'storeAnalogFxoChannelGroup',
  url: Paraxip.baseUrl + "call/getChannelGroups",
  root: 'data',
  fields: ['id']
});

var assignAnalogFxoInterfaceWindow = new Ext.Window({
  title       : 'Assign Analog Interface to Fxo Config',
  layout      : 'fit',
  layoutConfig: {
      autoWidth : false
  },
  width       : 600,
  height      : 500,
  closeAction : 'hide',
  resizable   : false,
  plain       : true,
  id          : "assignAnalogFxoInterfaceWindow"
});

var analogFxoInterfaceGrid = new Ext.grid.EditorGridPanel({
  title       : 'Analog Interfaces assigned to config',
  id          : 'analogFxoInterfaceGrid',
  autoScroll  : true,
  width       : 500,
  height      : 400,
  enableColumnMove : false,
  selModel    : new Ext.grid.RowSelectionModel(),
  store       : new Ext.data.JsonStore({
    id: 'storeAnalogFxoInterfaceGrid',
    url: Paraxip.baseUrl + "call/getAssignedFxoInterfaces",
    root: 'data',
    fields: [
      {name: 'boardId'},
      {name: 'interfaceId'},
      {name: 'channelGroupID'},
      {name: 'nodeId'} //not displayed
    ]
  }),
  columns     :[
    {
      header: "Board ID",
      width: 95,
      sortable: true,
      dataIndex: 'boardId',
      editor: new Ext.form.ComboBox({
        forceSelection  : true,
        triggerAction   : 'all',
        mode            : 'local',
        selectOnFocus   : true,
        valueField      : 'id',
        displayField    : 'id',
        store           : storeAnalogFxoBoards,
        listClass       : 'x-combo-list-small',
        validator       : Paraxip.config.callControl.validateID
      })
    },
    {
      header: "Interface ID",
      width: 95,
      sortable: true,
      dataIndex: 'interfaceId',
      editor: new Ext.form.ComboBox({
        forceSelection  : true,
        triggerAction   : 'all',
        mode            : 'local',
        selectOnFocus   : true,
        valueField      : 'id',
        displayField    : 'id',
        store           : storeAnalogFxoInterfaces,
        listClass       : 'x-combo-list-small',
        validator       : Paraxip.config.callControl.validateID
      })
    },
    {
      header: "Channel Group ID",
      width: 105,
      sortable: true,
      dataIndex: 'channelGroupID',
      editor: new Ext.form.ComboBox({
        forceSelection  : true,
        triggerAction   : 'all',
        mode            : 'local',
        selectOnFocus   : true,
        valueField      : 'id',
        displayField    : 'id',
        store           : storeAnalogFxoChannelGroup,
        listClass       : 'x-combo-list-small'
      })

    }
  ],
  listeners       : {
    'beforeedit':function(e){
      if( e.field == "interfaceId" ) {
        var colModel = e.grid.getColumnModel();
        var selectedRecord = e.grid.getSelectionModel().getSelected();
        var interfaceCellEditor = colModel.getCellEditor(e.column,e.row);
        interfaceCellEditor.field.store.load({
          url: Paraxip.baseUrl + "call/getAnalogFxoInterfaces",
          disableCaching: false,
          params: {'id':selectedRecord.data.boardId}
        });
      }
    }
    ,'afteredit':function(e){
      if( e.field == "boardId" ) {
        var colModel = e.grid.getColumnModel();
        var selectedRecord = e.grid.getSelectionModel().getSelected();
        var interfaceCellEditor = colModel.getCellEditor(e.column+1,e.row);
        interfaceCellEditor.field.store.load({
          url: Paraxip.baseUrl + "call/getAnalogFxoInterfaces",
          disableCaching: false,
          params: {'id':selectedRecord.data.boardId}
        });
      }
    }
  },
  viewConfig:{forceFit:true}
});

analogFxoInterfaceGrid.addButton("Add Analog Interface"
                                  ,addAnalogFxoInterface
                                  ,analogFxoInterfaceGrid);
analogFxoInterfaceGrid.addButton("Delete"
                                  ,deleteAnalogInterface
                                  ,analogFxoInterfaceGrid);
analogFxoInterfaceGrid.addButton("Save"
                                  ,applyToAnalogConfig
                                  ,analogFxoInterfaceGrid);
analogFxoInterfaceGrid.addButton("Cancel",function(){
  assignAnalogFxoInterfaceWindow.hide();
});

assignAnalogFxoInterfaceWindow.add(analogFxoInterfaceGrid);

//------------------------------------------------
// Analog FXS config form
// -----------------------------------------------
function saveFxsConfig(e){
  var owner = e.ownerCt;
  var form = owner.getForm();
  var el = form.getEl();

  el.dom.action = Paraxip.baseUrl + "call/saveFxsConfig";
  el.dom.method = 'POST';

  var fxsConfig = owner.getItemsValueAsJson();

  form.submit({
    params:{
      'frmAnalogFxs': fxsConfig
    },
    success: function(){
      Paraxip.isConfigOutOfSynch(true);
      reloadTree(true);
    },
    failure: Paraxip.failureMessage
  });
};

var frmAnalogFxs = new Ext.FormPanel({
  title         : "Analog FXS config"
  ,frame        : true
  ,header       : false
  ,id           : "frmAnalogFxs"
  ,autoScroll   : true
  ,layout       : 'accordion'
  ,layoutConfig : {
    activeItem  : 0
  }
  ,defaults     : {
    layout      : 'form'
    ,style      : 'text-align:left;'
    ,labelWidth : 150
    ,bodyStyle  : 'padding:5px;'
    ,width      : 500
    ,autoHeight : true
    ,allowBlank : false
    ,fill       : true
  }
  ,buttonAlign  : "right"
  ,buttons      : [{
    id        : "Paraxip.config.callControl.fxsAssignedInterfaceButton"
    ,text     : 'Assigned Interface(s)'
    ,listeners: {
      'click':function(){
        fxsConfigID = Paraxip.config.callControl.fxs.ID.getValue();
        analogFxsInterfaceGrid.getStore().load({
          url: Paraxip.baseUrl + "call/getAssignedFxsInterfaces"
          ,params: {'id':fxsConfigID}
        });
        storeAnalogFxsBoards.load({
          url: Paraxip.baseUrl + "call/getAnalogFxsBoards"
          ,params: {'id':fxsConfigID}
        });
        assignAnalogFxsInterfaceWindow.show();
      }
    }
  },{
    id        : "Paraxip.config.callControl.saveFxsConfigButton"
    ,text     : 'Save'
    ,listeners: {'click':saveFxsConfig}
  },{
    id        : "Paraxip.config.callControl.cancelSaveConfigButton"
    ,text     : 'Cancel'
    ,listeners: {'click':callControlCancel}
  }]
});

// FXS config generic params panel
Paraxip.config.callControl.fxs.configPanel = new ParaxipPanelWithHelp({
  id      : 'Paraxip.config.callControl.fxs.configPanel'
  ,name   : 'fxsConfig'
  ,title  : 'Configuration Info'
  ,items  : [{
    xtype     : 'textfield'
    ,name     : 'nodeId'
    ,hidden   : true
    ,hideLabel: true
  }]
  ,helpText: "analogFxsConfiganalogFxsConfig"
});

Paraxip.config.callControl.fxs.inboundCallHandlingPanel =
    new ParaxipPanelWithHelp({
  id        : 'Paraxip.config.callControl.fxs.inboundCallHandlingPanel'
  ,name     : 'inboundCallHandling'
  ,title    : 'Inbound Call - Handling Info'
  ,helpText : "fxsInboundCallParametersinboundCallHandling"
});

Paraxip.config.callControl.fxs.hotlineDialingPanel = new ParaxipPanelWithHelp({
  id        : 'Paraxip.config.callControl.fxs.hotlineDialingPanel'
  ,name     : 'hotlineDialing'
  ,title    : 'Hotline Dialing - Info'
  ,helpText : "fxsHotlineDialingParametershotlineDialing"
});

Paraxip.config.callControl.fxs.disconnectAnswerPanel =
  new ParaxipPanelWithHelp({
    id        : 'Paraxip.config.callControl.fxs.disconnectAnswerPanel'
    ,name     : 'disconnectAnswerNotification'
    ,title    : 'Disconnect and Answer Notification'
    ,helpText : "fxsDisconnectAnswerNotificationParametersdisconnect" +
                "AnswerNotification"
  });

Paraxip.config.callControl.fxs.ID = createIdText({
  id        : 'Paraxip.config.callControl.fxs.ID'
  ,helpText : 'analogFxoConfiganalogFxoConfigID'
});

Paraxip.config.callControl.fxs.displayName = createNameText({
  id        : 'Paraxip.config.callControl.fxs.displayName'
  ,helpText : 'analogFxoConfiganalogFxoConfigdisplayName'
});

// regretTime
Paraxip.config.callControl.fxs.regretTime = new Ext.form.NumberField({
  id              : 'Paraxip.config.callControl.fxs.regretTime'
  ,fieldLabel     : "Regret Time (ms)"
  ,name           : "regretTimeMs"
  ,helpText       : "analogFxsConfiganalogFxsConfigregretTimeMs"
  ,allowNegative  : false
  ,allowBlank     : false
  ,allowDecimals  : false
  ,width          : 252
});

// oosBehavior
Paraxip.config.callControl.fxs.oosBehavior = new Ext.form.ComboBox({
  id                : 'Paraxip.config.callControl.fxs.oosBehavior'
  ,name             : 'oosBehavior'
  ,fieldLabel       : "Out-Of-Service Behavior"
  ,helpText         : "analogFxsConfiganalogFxsConfigoosBehavior"
  ,store            : ['REORDER','NONE']
  ,value            : "REORDER"
  ,forceSelection   : true
  ,triggerAction    : 'all'
  ,mode             : 'local'
  ,typeAhead        : true
  ,selectOnFocus    : true
  ,width            : 145
  ,minListWidth     : 145
});

// minExpectedDigits
Paraxip.config.callControl.fxs.minExpectedDigits = new Ext.form.NumberField({
  id              : 'Paraxip.config.callControl.fxs.minExpectedDigits'
  ,fieldLabel     : "Minimum Expected Digits"
  ,name           : "minExpectedDigits"
  ,helpText       : "fxsInboundCallParametersinboundCallHandling" +
                    "minExpectedDigits"
  ,allowNegative  : false
  ,allowBlank     : false
  ,allowDecimals  : false
  ,width          : 252
});

// busyToneDuration
Paraxip.config.callControl.fxs.busyToneDuration = new Ext.form.NumberField({
  id              : 'Paraxip.config.callControl.fxs.busyToneDuration'
  ,fieldLabel     : "Busy Tone Duration (ms)"
  ,name           : "busyToneDurationMs"
  ,helpText       : "fxsInboundCallParametersinboundCallHandling" +
                    "busyToneDurationMs"
  ,allowNegative  : false
  ,allowBlank     : false
  ,allowDecimals  : false
  ,width          : 252
});

// interDigitTimeout
Paraxip.config.callControl.fxs.interDigitTimeout = new Ext.form.NumberField({
  id              : 'Paraxip.config.callControl.fxs.interDigitTimeout'
  ,fieldLabel     : "Inter Digit Timeout (ms)"
  ,name           : "interDigitTimeoutMs"
  ,helpText       : "fxsInboundCallParametersinboundCallHandling" +
                    "interDigitTimeoutMs"
  ,allowNegative  : false
  ,allowBlank     : false
  ,width          : 252
});

// dialToneDuration
Paraxip.config.callControl.fxs.dialToneDuration = new Ext.form.NumberField({
  id              : "Paraxip.config.callControl.fxs.dialToneDuration"
  ,fieldLabel     : "Dial Tone Duration (ms)"
  ,name           : "dialToneDurationMs"
  ,helpText       : "fxsInboundCallParametersinboundCallHandling" +
                    "dialToneDurationMs"
  ,allowNegative  : false
  ,allowBlank     : false
  ,allowDecimals  : false
  ,width          : 252
});

// relayEarlyMedia
Paraxip.config.callControl.fxs.relayEarlyMedia =
    new Ext.form.ComboBox({
  id              : 'Paraxip.config.callControl.fxs.relayEarlyMedia'
  ,name           : 'relayEarlyMedia'
  ,fieldLabel     : "Relay Early Media"
  ,helpText       : "fxsRelayEarlyMediaParameterrelayEarlyMedia"
  ,store          : [ 'ENABLED', 'DISABLED' ]
  ,forceSelection : true
  ,triggerAction  : 'all'
  ,mode           : 'local'
  ,typeAhead      : true
  ,selectOnFocus  : true
  ,width          : 200
  ,minListWidth   : 200
});

// hotlineDialingActive
Paraxip.config.callControl.fxs.hotlineDialingActive = new Ext.form.ComboBox({
  id              : 'Paraxip.config.callControl.fxs.hotlineDialingActive'
  ,name           : 'hotlineDialingActive'
  ,fieldLabel     : "Hotline Dialing Active"
  ,helpText       : "fxsHotlineDialingParametershotlineDialingenable"
  ,store          : ['YES','NO']
  ,forceSelection : true
  ,triggerAction  : 'all'
  ,mode           : 'local'
  ,typeAhead      : true
  ,selectOnFocus  : true
  ,width          : 145
  ,minListWidth   : 145
  ,listeners      : {
    'select':isHotlineDialingEnable
  }
});

Paraxip.config.callControl.fxs.validateDelayBeforeDialing = function() {
  if( Paraxip.config.callControl.fxs.hotlineDelayBeforeDialing.getValue() <
      Paraxip.config.callControl.fxs.dialToneDuration.getValue() ) {
    return true;
  } else {
    return "The delay before dialing value must be lower than the dial " +
           "tone duration."
  }
}

// delayBeforeDialing
Paraxip.config.callControl.fxs.hotlineDelayBeforeDialing =
    new Ext.form.NumberField({
  id              : 'Paraxip.config.callControl.fxs.hotlineDelayBeforeDialing'
  ,fieldLabel     : "Delay Before Dialing (ms)"
  ,name           : "delayBeforeDialingMs"
  ,helpText       : "fxsHotlineDialingParametershotlineDialing" +
                    "delayBeforeDialingMs"
  ,allowNegative  : false
  ,allowBlank     : false
  ,allowDecimals  : false
  ,validator      : Paraxip.config.callControl.fxs.validateDelayBeforeDialing
  ,width          : 252
});

// dialingNumber
Paraxip.config.callControl.fxs.hotlineDialingNumber = new Ext.form.TextField({
  id            : 'Paraxip.config.callControl.fxs.hotlineDialingNumber'
  ,fieldLabel   : "Dialing Number"
  ,name         : "dialingNumber"
  ,helpText     : "fxsHotlineDialingParametershotlineDialingdialingNumber"
  ,allowBlank   : false
  ,width        : 252
});

Paraxip.config.callControl.fxs.disconnectAnswerNotificationMode =
  new Ext.form.ComboBox({
    id              : 'Paraxip.config.callControl.fxs.disconnectAnswerNotificationMode'
    ,name           : "disconnectAnswerMode"
    ,fieldLabel     : "Mode"
    ,helpText       : "fxsDisconnectAnswerNotificationParametersdisconnect" +
                      "AnswerNotification"
    ,store          : ['NONE','POLARITY-REVERSAL'/*,'CURRENT-DISCONNECT'*/]
    ,value          : "POLARITY-REVERSAL"
    ,forceSelection : true
    ,triggerAction  : 'all'
    ,mode           : 'local'
    ,typeAhead      : true
    ,selectOnFocus  : true
    ,width          : 145
    ,minListWidth   : 145
});

// TODO : add currentDisconnectDurationMs field when the current-disconnect
// feature will be available

// Add the attributes to their respective panels and add the panels to the main
// frame.

// Config Panel
Paraxip.config.callControl.fxs.configPanel.add(
    Paraxip.config.callControl.fxs.ID);
Paraxip.config.callControl.fxs.configPanel.add(
    Paraxip.config.callControl.fxs.displayName);
Paraxip.config.callControl.fxs.configPanel.add(
    Paraxip.config.callControl.fxs.regretTime);
Paraxip.config.callControl.fxs.configPanel.add(
    Paraxip.config.callControl.fxs.oosBehavior);

frmAnalogFxs.add(Paraxip.config.callControl.fxs.configPanel);

// Inbound Call Handling Panel
Paraxip.config.callControl.fxs.inboundCallHandlingPanel.add(
    Paraxip.config.callControl.fxs.minExpectedDigits);
Paraxip.config.callControl.fxs.inboundCallHandlingPanel.add(
    Paraxip.config.callControl.fxs.interDigitTimeout);
Paraxip.config.callControl.fxs.inboundCallHandlingPanel.add(
    Paraxip.config.callControl.fxs.busyToneDuration);
Paraxip.config.callControl.fxs.inboundCallHandlingPanel.add(
    Paraxip.config.callControl.fxs.dialToneDuration);
Paraxip.config.callControl.fxs.inboundCallHandlingPanel.add(
    Paraxip.config.callControl.fxs.relayEarlyMedia);

frmAnalogFxs.add(Paraxip.config.callControl.fxs.inboundCallHandlingPanel);

// Hotline Dialing Panel
Paraxip.config.callControl.fxs.hotlineDialingPanel.add(
    Paraxip.config.callControl.fxs.hotlineDialingActive);
Paraxip.config.callControl.fxs.hotlineDialingPanel.add(
    Paraxip.config.callControl.fxs.hotlineDelayBeforeDialing);
Paraxip.config.callControl.fxs.hotlineDialingPanel.add(
    Paraxip.config.callControl.fxs.hotlineDialingNumber);

frmAnalogFxs.add(Paraxip.config.callControl.fxs.hotlineDialingPanel);

// Disconnect Answer Notification Panel
Paraxip.config.callControl.fxs.disconnectAnswerPanel.add(
    Paraxip.config.callControl.fxs.disconnectAnswerNotificationMode);

frmAnalogFxs.add(Paraxip.config.callControl.fxs.disconnectAnswerPanel);

function addAnalogFxsInterface(e){
  var form = e.ownerCt;

  if( storeAnalogFxsBoards.data.length < 1 ) {
    Ext.Msg.alert(
      "No Analog boards available","There is no analog boards in your " +
      "configuration that contains FXS interfaces. Please add a board in " +
      "the board section if you want to assign a board to this config.");
    return;
  }

  var firstBoardId = storeAnalogFxsBoards.getAt(0).get('id');
  storeAnalogFxsInterfaces.load({
    url: Paraxip.baseUrl + "call/getAnalogFxsInterfaces",
    disableCaching: false,
    params: {'id':firstBoardId},
    callback: function(result){
      if( result.length > 0 ) {
        var newRecord = new Ext.data.Record({
          'boardId':firstBoardId,
          'interfaceId':storeAnalogFxsInterfaces.getAt(0).get('id'),
          'channelGroupID':storeAnalogFxsInterfaces.getAt(0).get('phoneNumber')
        });
        analogFxsInterfaceGrid.getStore().add(newRecord);
        analogFxsInterfaceGrid.expand();
      }
    }
  });
};

var fxsConfigID = '';

// Store of the boards that have fxs interfaces
var storeAnalogFxsBoards = new Ext.data.JsonStore({
  id      : 'storeAnalogFxsBoards'
  ,url    : Paraxip.baseUrl + "call/getAnalogFxsBoards"
  ,root   : 'data'
  ,fields : ['id','displayName','nodeId']
});

// Store of the fxs interfaces for a selected board
var storeAnalogFxsInterfaces = new Ext.data.JsonStore({
  id      : 'storeAnalogFxsInterfaces'
  ,url    : Paraxip.baseUrl + "call/getAnalogFxsInterfaces"
  ,root   : 'data'
  ,fields : ['id','displayName','phoneNumber']
});

var assignAnalogFxsInterfaceWindow = new Ext.Window({
  title         : 'Assign Analog Interface to Fxs Config'
  ,layout       : 'fit'
  ,layoutConfig : {
    autoWidth : false
  }
  ,width        : 600
  ,height       : 500
  ,closeAction  : 'hide'
  ,resizable    : false
  ,plain        : true
  ,id           : "assignAnalogFxsInterfaceWindow"
});

var analogFxsInterfaceGrid = new Ext.grid.EditorGridPanel({
  title       : 'Analog Interfaces assigned to config',
  id          : 'analogFxsInterfaceGrid',
  autoScroll  : true,
  width       : 500,
  height      : 400,
  enableColumnMove : false,
  selModel    : new Ext.grid.RowSelectionModel(),
  store       : new Ext.data.JsonStore({
    id: 'storeAnalogFxsInterfaceGrid',
    url: Paraxip.baseUrl + "call/getAssignedFxsInterfaces",
    root: 'data',
    fields: [
      {name: 'boardId'},
      {name: 'interfaceId'},
      {name: 'channelGroupID'},
      {name: 'nodeId'} //not displayed
    ]
  }),
  columns     :[
    {
      header: "Board ID",
      width: 95,
      sortable: true,
      dataIndex: 'boardId',
      editor: new Ext.form.ComboBox({
        forceSelection  : true,
        triggerAction   : 'all',
        mode            : 'local',
        selectOnFocus   : true,
        valueField      : 'id',
        displayField    : 'id',
        store           : storeAnalogFxsBoards,
        listClass       : 'x-combo-list-small',
        validator       : Paraxip.config.callControl.validateID
      })
    },
    {
      header: "Interface ID",
      width: 95,
      sortable: true,
      dataIndex: 'interfaceId',
      editor: new Ext.form.ComboBox({
        forceSelection  : true,
        triggerAction   : 'all',
        mode            : 'local',
        selectOnFocus   : true,
        valueField      : 'id',
        displayField    : 'id',
        store           : storeAnalogFxsInterfaces,
        listClass       : 'x-combo-list-small',
        validator       : Paraxip.config.callControl.validateID
      })
    },
    {
      header: "Channel Group ID",
      width: 105,
      sortable: true,
      dataIndex: 'channelGroupID',
      editor: new Ext.form.TextField({
        readOnly: true
      })
    }
  ],
  listeners       : {
    'beforeedit':function(e){
      if( e.field == "interfaceId" ) {
        var colModel = e.grid.getColumnModel();
        var selectedRecord = e.grid.getSelectionModel().getSelected();
        var interfaceCellEditor = colModel.getCellEditor(e.column,e.row);
        interfaceCellEditor.field.store.load({
          url: Paraxip.baseUrl + "call/getAnalogFxsInterfaces",
          disableCaching: false,
          params: {'id':selectedRecord.data.boardId}
        });
      }
    }
    ,'afteredit':function(e){
      if( e.field == "boardId" ) {
        var colModel = e.grid.getColumnModel();
        var selectedRecord = e.grid.getSelectionModel().getSelected();
        var interfaceCellEditor = colModel.getCellEditor(e.column+1,e.row);
        interfaceCellEditor.field.store.load({
          url: Paraxip.baseUrl + "call/getAnalogFxsInterfaces",
          disableCaching: false,
          params: {'id':selectedRecord.data.boardId},
          callback: function(){
            storeAnalogFxsInterfaces.each(function(record){
              if( record.data.id == e.record.data.interfaceId ) {
                e.record.set('channelGroupID', record.data.phoneNumber);
              }
            })
          }
        });
      } else if( e.field == "interfaceId" ) {
        storeAnalogFxsInterfaces.each(function(record){
          if( record.data.id == e.value ) {
            e.record.set('channelGroupID', record.data.phoneNumber);
          }
        })
      }
    }
  },
  viewConfig:{forceFit:true}
});

analogFxsInterfaceGrid.addButton("Add Analog Interface"
                                  ,addAnalogFxsInterface
                                  ,analogFxsInterfaceGrid);
analogFxsInterfaceGrid.addButton("Delete"
                                  ,deleteAnalogInterface
                                  ,analogFxsInterfaceGrid);
analogFxsInterfaceGrid.addButton("Save"
                                  ,applyToAnalogConfig
                                  ,analogFxsInterfaceGrid);
analogFxsInterfaceGrid.addButton("Cancel",function(){
  assignAnalogFxsInterfaceWindow.hide();
});

assignAnalogFxsInterfaceWindow.add(analogFxsInterfaceGrid);
