/**
 * @fileBrowser.js
 * @author: Jean-Philippe Boily
 * date: Mon Mar 23 2009
 *
 * Description: 
 *
 *  Create an instance of the FileTreePanel class and Panels
 *  to create the File Browser of the WebUI.
 *
 *  Copyright (C) 2009 Paraxip Technologies.
 *  All Rights Reserved.
 */

Ext.namespace("Paraxip.fileBrowser");

// Always know which is the last node that was clicked to
// handle the callbacks.
Paraxip.fileBrowser.lastClickedFile = new Ext.tree.TreeNode();

// Maximum size in bytes of a file for wich we will load its
// content in the browser.
Paraxip.fileBrowser.maximumLoadableFileSize = 256000; 

// Maximum size in bytes where the file is too big to be
// displayed in the browser.
Paraxip.fileBrowser.maximumUnloadableFileSize = 1048576; 

// When more than this number of files in a directory, browser
// is not responsive for a while.
Paraxip.fileBrowser.dirItemsLimit = 200;

// Make this variable global to keep trace of its value
allowedToContinue = null;

// When a file has an extension contained in that array, the
// file will be downloaded on a simple click instead of trying
// to open it in the TextArea like other files.
Paraxip.fileBrowser.DownlaodOnClickExtensionArray = new Array(
  "pdf",
  "zip"
);

Paraxip.fileBrowser.fileTree = new Ext.ux.FileTreePanel({
  id              : 'Paraxip.fileBrowser.fileTree',
  region          : 'west',
  width           : 225,
  minSize         : 250,
  maxSize         : 250,
  enableUpload    : false,
  enableNewDir    : false,
  enableDelete    : false,
  enableRename    : false,
  enableDD        : false,
  enableProgress  : false,
  autoScroll      : true,
  openMode        : 'download',
  //rootVisible     : false,
  rootText        : 'root',
  url             : Paraxip.baseUrl + "fileBrowser/"
  //url             : '/fileBrowser/'
});

Paraxip.fileBrowser.getFileContent = function(node){
  if( node.attributes.leaf ) {
    Paraxip.fileBrowser.lastClickedFile = node;

    if( Paraxip.isItemInList(Paraxip.fileBrowser.DownlaodOnClickExtensionArray,
                             node.attributes.extension ) ) {
      node.fireEvent("dblclick",node);
      return;
    }
    Ext.Ajax.request({
      url: Paraxip.baseUrl + "fileBrowser/getFileSize",
      params: {
        'filePath':node.getPath('text')
      },
      success: function(result){
        var response = Ext.util.JSON.decode(result.responseText);
        if( response.fileSize>Paraxip.fileBrowser.maximumUnloadableFileSize ) {
          Ext.Msg.alert(
            "File size over 1MB",
            "The file you are about to open is over 1MB wich too big to be " +
            "open in the browser. To see its content, please download the " +
            "file on your computer."
          );
        } else if(
            response.fileSize<=Paraxip.fileBrowser.maximumUnloadableFileSize &&
            response.fileSize >= Paraxip.fileBrowser.maximumLoadableFileSize) {
          Ext.Msg.confirm(
            "File size over 250KB",
            "The file you are about to open is over 250KB wichcould cause a " +
            "lag in the UI. Do you want to proceed?",
            Paraxip.fileBrowser.showFileInBrowser
          );
        } else {
          Paraxip.fileBrowser.showFileInBrowser("yes");
        }
      }
    })
    
  }
}

Paraxip.fileBrowser.getNumItemsInDir = function(node){
  if( allowedToContinue != null ) {
    var tempBool = allowedToContinue;
    allowedToContinue = null;
    return tempBool;
  }
  Ext.Ajax.request({
    url: Paraxip.baseUrl + "fileBrowser/getNumItemsInDir",
    params: {
      'filePath':node.getPath('text')
    },
    success: function(result){
      var response = Ext.util.JSON.decode(result.responseText);
      if( response.numOfItems > Paraxip.fileBrowser.dirItemsLimit ) {
        Ext.Msg.confirm(
          "Directory has many files",
          "There is " + response.numOfItems + " files in the directoy which " +
          "will make the browser unresponsive for a while. Do you want to " +
          "continue?",
          function(answer){
            if( answer == "yes" ) {
              allowedToContinue = true;
            } else {
              allowedToContinue = false;
            }
            node.expand();
          }
        );
      } else {
        allowedToContinue = true;
        node.expand();
      }
    }
  });

  return false;
}

Paraxip.fileBrowser.showFileInBrowser = function(openFile){
  if( openFile == "yes" ) {
    Ext.Ajax.request({
      url: Paraxip.baseUrl + "fileBrowser/showFileInBrowser",
      params: {
        'filePath':Paraxip.fileBrowser.lastClickedFile.getPath('text')
      },
      success: function(result){
        var response = Ext.util.JSON.decode(result.responseText);
        Paraxip.fileBrowser.saveFile.setDisabled(
          Paraxip.fileBrowser.lastClickedFile.disabled);
        Paraxip.fileBrowser.clearChanges.setDisabled(
          Paraxip.fileBrowser.lastClickedFile.disabled);
        Paraxip.fileBrowser.fileEditor.setValue(response.fileText);
        Ext.getCmp("Paraxip.fileBrowser.description").setText(
          "<html><p align='center'>This section of the webUI allows you to " +
          "read and edit configuration and logging files. Grayed out file " +
          "name in the tree means a read only files. It is possible to " +
          "download files by right-clicking on a file in the tree panel." +
          "<br/><br/><b>File: " +
          Paraxip.fileBrowser.lastClickedFile.getPath('text') + 
          "</b></p></html>", false
        );
      },
      failure: Paraxip.failureMessage
    });
  }
}

Paraxip.fileBrowser.saveFileContent = function(){
  Ext.Ajax.request({
    url: Paraxip.baseUrl + "fileBrowser/saveFileChanges",
    params: {
      'filePath':Paraxip.fileBrowser.lastClickedFile.getPath('text'),
      'fileContent':Paraxip.fileBrowser.fileEditor.getValue()
    },
    success: function(){
       Ext.Msg.alert(Paraxip.fileBrowser.lastClickedFile.text + " saved!", 
                     "File saved successfully.");
    },
    failure: Paraxip.failureMessage
  });
}

Paraxip.fileBrowser.fileTree.on('click', Paraxip.fileBrowser.getFileContent);
Paraxip.fileBrowser.fileTree.on(
  'beforeexpandnode', Paraxip.fileBrowser.getNumItemsInDir);

Paraxip.fileBrowser.saveFile = new Ext.Button({
  id        : 'Paraxip.fileBrowser.saveFile',
  text      : 'Save Changes',
  minWidth  : 88,
  disabled  : true,
  listeners : {
    'click':Paraxip.fileBrowser.saveFileContent
  }
});

Paraxip.fileBrowser.clearChanges = new Ext.Button({
  id        : 'Paraxip.fileBrowser.clearChanges',
  text      : 'Clear Changes',
  minWidth  : 88,
  disabled  : true,
  listeners : {
    'click':function(){
      Paraxip.fileBrowser.getFileContent(Paraxip.fileBrowser.lastClickedFile);
    }
  }
});

Paraxip.fileBrowser.fileEditor = new Ext.form.TextArea({
  id      : 'Paraxip.fileBrowser.fileEditor'
  ,width  : 745
  ,height : 410
  ,style  : 'font-family:"Courrier New", monospace;'
});

Paraxip.fileBrowser.fileEditorForm = new Ext.form.FormPanel({
  id      : 'Paraxip.fileBrowser.fileEditorForm',
  region  : 'center',
  items   : [{
    xtype:'label',
    autoHeight: true,
    style: "font-size: 12",
    id:'Paraxip.fileBrowser.description',
    html: "<html><p align='center'>This section of the webUI allows you to " +
          "read and edit configuration and logging files. Grayed out file " +
          "name in the tree means a read only files. It is possible to " +
          "download files by right-clicking on a file in the tree panel." +
          "<br/><br/><br/></p></html>"
  },
    Paraxip.fileBrowser.fileEditor
  ],
  buttons : [
    Paraxip.fileBrowser.saveFile,
    Paraxip.fileBrowser.clearChanges
  ]
});

Paraxip.fileBrowser.fileBrowserTabPanel = new Ext.Panel({
  id        : 'Paraxip.fileBrowser.fileBrowserTabPanel',
  layout    : 'border',
  defaults  : {
    style: 'text-align:left',
    layout: 'fit'
  },
  items     : [
    Paraxip.fileBrowser.fileTree,
    Paraxip.fileBrowser.fileEditorForm
  ]
});

