// Allows the use of the Firebug Console API statements 
//  (http://getfirebug.com/console.html) whether Firebug is available or not  
// Exposed thru the variable "Paraxip.console" instead of "console" (done to 
//  protect against imminent naming changes - see 
//  http://code.google.com/p/fbug/issues/detail?id=109), avoid clashes with 
//  other browser logging lib (ex Safari), and to allow extension of the API. 
// Ex. use Paraxip.console.info('text') instead of console.info('text')

// Make sure "Paraxip" object is initialized
if (!("Paraxip" in window)) {
  Paraxip = {};
}

// Check if our console was already initialized
if (!("console" in window.Paraxip)) {

  // window.loadFirebugConsole() defined by Firebug 1.2.1, and needed for Firebug
  // in FF 2 to install the console object
  if ("loadFirebugConsole" in window) {
    window.loadFirebugConsole();
  }

  // See if "window.console" object is installed
  if ("console" in window) {

    // Is Firebug the provider of the "console" object?
    // console.firebug is a string holding Firebug version
    if ("firebug" in console) {

      console.log('Paraxip console using Firebug %s', console.firebug)

      // assert() method is broken in Firebug 1.2, see
      //   http://code.google.com/p/fbug/issues/detail?id=998
      if (/^1\.2/.test(console.firebug)) {
        // Replacement for assert
        console.assert = function(expression) {
          if (!expression) {
            console.error('Assertion failed');
            // Pass remaining arguments to console.error
            console.error(Array.prototype.slice.call(arguments).slice(1));
            // Break into the debugger
            debugger;
          }
        }
      }

      // Is Companion.JS the provider of the "console" object?
    } else if ("provider" in console && console.provider == "Companion.JS") {

      console.log('Paraxip console using Companion.JS %s', console.currentVersion)

    } else {

      console.info('Paraxip console using unknown Console API provider')

    }

    // Set Paraxip.console as an alias to "console"
    Paraxip.console = console;

  } else {

    // No "console" detected - stub the console API in "Paraxip.console"
    var functionNames = ["log", "debug", "info", "warn", "error", "assert",
      "dir", "dirxml", "group", "groupEnd", "time", "timeEnd", "count",
      "trace", "profile", "profileEnd"];

    Paraxip.console = {};
    for (var i in functionNames) {
      Paraxip.console[functionNames[i]] = function() {
      };
    }

  }

}

// ExtJS-specific

// Trace all fired events for a given Ext JS Component
Paraxip.console.traceEventsForComponent = function(component) {
  for (event in component.events) {
    var handler = function() {
          Paraxip.console.info(this);
          return true;
        };
    // createDelegate (Extjs) will set given context as "this" in callback
    var context = {id: component.id, eventName: event}
    component.on(event, handler.createDelegate(context));
  }
};
