Ext.namespace("Paraxip",
              "Paraxip.status",
              "Paraxip.status.system",
              "Paraxip.status.controls",
              "Paraxip.status.channels");

// Defines how many time we have to ping the gateway to see if the application
// is started.
Paraxip.pingGatewayRepeatCount = 25;

if( pingGW == undefined ){
  var pingGW = false;
}

var gwStatusRefreshTime = 30;
var refreshGWStatusTask = undefined;


/**
 * Stops the automatic refresh of the grids in Status&Controls. Might be useful
 * in development mode only.
 */
Paraxip.status.controls.stopAutomaticRefreshTime = function(){
  if( Paraxip.status.system.errorGrid != undefined ) {
    Paraxip.status.system.errorGrid.getStore().stopAutoRefresh();
  }
  if( Paraxip.status.system.statsGrid != undefined ) {
    Paraxip.status.system.statsGrid.getStore().stopAutoRefresh();
  }
  if( Paraxip.status.channels.channelGrid != undefined ) {
    Paraxip.status.channels.channelGrid.getStore().stopAutoRefresh();
  }
};

Paraxip.status.controls.startGwButton = new Ext.Button({
  id: 'Paraxip.status.controls.startGwButton',
  text: 'Start Gateway',
  disabled: true,
  minWidth: 200
});

Paraxip.status.controls.stopGwButton = new Ext.Button({
  id: 'Paraxip.status.controls.stopGwButton',
  text: 'Stop Gateway',
  disabled: true,
  minWidth: 200
});

Paraxip.status.controls.restartGwButton = new Ext.Button({
  id: 'Paraxip.status.controls.restartGwButton',
  text: 'Restart Gateway',
  disabled: true,
  minWidth: 200
});

Paraxip.status.controls.refreshGWStatus = new Ext.Button({
  id: 'Paraxip.status.controls.refreshGWStatus',
  text: 'Refresh Gateway Status',
  minWidth: 200,
  listeners: {
    'click':Paraxip.isGwStartedCall
  }
});

Paraxip.status.controls.startGwButton.on('click', function() {
  Paraxip.changeGwStatusBar(Paraxip.serviceStartingStatus, true, true);
  Paraxip.showProgressBar("Starting the gateway service...");
  Paraxip.disableItems(true);
  Ext.Ajax.request({
    url: Paraxip.baseUrl + "status/startGatewayButton",
    disableCaching: true,
    //1 minute timeout to prevent error when gateway takes a long time to start
    timeout: startGwTimeout,
    params: {js:Ext.util.JSON.encode('dummy param')},
    success: function(result){
      var response = Ext.util.JSON.decode(result.responseText);
      if (response.value) {
        Paraxip.serviceStarted = true;
        Paraxip.changeGwStatusBar(
          Paraxip.initializingApplicationStatus, true, true);
        Paraxip.updateProgressBarText(
          "Service started. Initializing Gateway Application...");
        Paraxip.pingGatewayRepeatCount = 25;
        Ext.TaskMgr.start(Paraxip.status.controls.pingGwTask);
      } else {
        Paraxip.hideProgressBar();
        Paraxip.changeGwStatusBar(Paraxip.downStatus, false, true);
        Ext.Msg.alert(
          "Unable to start the gateway",
          "The gateway failed to start. Please be sure there is a valid " +
          "configuration file before launching the gateway. This can be " +
          "accomplished by running the Quick Setup."
        );
      }
    },
    failure: function(){
      Paraxip.hideProgressBar();
      Paraxip.changeGwStatusBar(Paraxip.downStatus, false, true);
      Paraxip.failureMessage();
    }
  });
});

Paraxip.status.controls.pingGwTask = {
  run: function(){Paraxip.status.controls.isApplicationStarted();},
  interval: 2000
};

Paraxip.status.controls.isApplicationStarted = function(){
  Paraxip.pingGatewayRepeatCount = Paraxip.pingGatewayRepeatCount - 1 ;
  if( Paraxip.pingGatewayRepeatCount == 0 ) {
    Ext.TaskMgr.stop(Paraxip.status.controls.pingGwTask);
    Paraxip.hideProgressBar();
    Paraxip.changeGwStatusBar(Paraxip.downStatus, false, true);
    Ext.Msg.alert(
      "Unable to start the gateway",
      "The gateway failed to start. Please be sure there is a valid " +
      "configuration file before launching the gateway. This can be " +
      "accomplished by running the Quick Setup."
    );
    return;
  }
  Ext.Ajax.request({
    url: Paraxip.baseUrl + "status/pingGateway",
    disableCaching: true,
    timeout: startGwTimeout,
    params: {js:Ext.util.JSON.encode('dummy param')},
    success: function(result){
      var response = Ext.util.JSON.decode(result.responseText);
      if (response.value) {
        Ext.TaskMgr.stop(Paraxip.status.controls.pingGwTask);
        Paraxip.isConfigOutOfSynch(false);
        Paraxip.isGwStartedCall();
        Paraxip.hideProgressBar();
      }
    },
    failure: function(){
      Ext.TaskMgr.stop(Paraxip.status.controls.pingGwTask);
      Paraxip.hideProgressBar();
      Paraxip.changeGwStatusBar(Paraxip.downStatus, false, true);
      Ext.Msg.alert(
        "Unable to start the gateway",
        "The gateway failed to start. Please be sure there is a valid " +
        "configuration file before launching the gateway. This can be " +
        "accomplished by running the Quick Setup."
      );
    }
  });
};

Paraxip.status.controls.stopGwButtonMethod = function(answer) {
  if ( answer == "yes" ) {
    Paraxip.serviceStarted = false;
    Paraxip.changeGwStatusBar(Paraxip.stoppingStatus, true, true);
    Paraxip.showProgressBar("Stopping the gateway...");
    Paraxip.disableItems(true);
    Ext.Ajax.request({
      url: Paraxip.baseUrl + "status/stopGatewayButton",
      disableCaching: true,
      params: {js:Ext.util.JSON.encode('dummy param')},
      success: function(result){
        var response = Ext.util.JSON.decode(result.responseText);
        if (response.value) {
          Paraxip.isGwStartedCall();
        } else {
          Paraxip.changeGwStatusBar(Paraxip.downStatus, false, true);
          Ext.Msg.alert(
            "Unable to stop the gateway",
            "Failed to stop the gateway. Gateway may already be stopped."
          );
        }
        Paraxip.hideProgressBar();
      },
      failure: function(){
        Paraxip.hideProgressBar();
        Paraxip.changeGwStatusBar(Paraxip.downStatus, false, true);
        Paraxip.failureMessage();
      }
    });
  }
};

Paraxip.status.controls.stopGwButton.on('click', function(){
  Ext.Msg.confirm("Stopping Gateway!","You are about to stop the gateway." +
                  "Are you sure you want to proceed?",
                  Paraxip.status.controls.stopGwButtonMethod);
});

Paraxip.status.controls.restartGwButtonMethod = function(answer) {
  if ( answer == "yes" ) {
    Paraxip.serviceStarted = false;
    Paraxip.gatewayStarted = false;
    Paraxip.changeGwStatusBar(Paraxip.serviceRestartingStatus, true, true);
    Paraxip.showProgressBar("Restarting the gateway...");
    Paraxip.disableItems(true);
    Ext.Ajax.request({
      url: Paraxip.baseUrl + "status/restartGateway",
      disableCaching: true,
      //1 minute timeout to prevent error when gateway takes a long time to restart
      timeout: startGwTimeout,
      params: {js:Ext.util.JSON.encode('dummy param')},
      success: function(result){
        var response = Ext.util.JSON.decode(result.responseText);
        if (response.success) {
          Paraxip.serviceStarted = true;
          Paraxip.changeGwStatusBar(
            Paraxip.initializingApplicationStatus, true, true);
          Paraxip.updateProgressBarText(
            "Service restarted. Initializing Gateway Application...");
          Paraxip.pingGatewayRepeatCount = 25;
          Ext.TaskMgr.start(Paraxip.status.controls.pingGwTask);
        } else {
          Paraxip.hideProgressBar();
          Paraxip.changeGwStatusBar(Paraxip.downStatus, false, true);
          Ext.Msg.alert(
            "Unable to start the gateway",
            "The gateway failed to start. Please be sure there is a valid " +
              "configuration file before launching the gateway. This can be " +
              "accomplished by running the Quick Setup."
          );
        }
      },
      failure: function(){
        Paraxip.hideProgressBar();
        Paraxip.changeGwStatusBar(Paraxip.downStatus, false, true);
        Paraxip.failureMessage();
      }
    });
  }
};

Paraxip.status.controls.restartGwButton.on('click', function(){
  Ext.Msg.confirm("Restarting Gateway!","You are about to restart the gateway." +
                  "Are you sure you want to proceed?",
                  Paraxip.status.controls.restartGwButtonMethod);
});

Paraxip.status.controls.autoRefreshTime = new Ext.form.NumberField({
  id: 'Paraxip.status.controls.autoRefreshTime'
  ,allowNegative: false
  ,allowDecimals: false
  ,allowBlank: false
  ,value: gwStatusRefreshTime
  ,minValue: 1
  ,anchor: '50%'
  ,fieldLabel: 'Automatic refresh interval (s)'
  ,listeners: {
    'render': function(){
      var elp = this.el.findParent('.x-form-element', 5, true);
      this.helpImage = elp.createChild({
        style : "width:16px;height:18px;position:relative;left:30px;" +
                "top:-23px;display:block;background:transparent " +
                "url(/static/images/help.gif) no-repeat 0 2px;" +
                "cursor:pointer!important;"
      });

      this.helpImage.alignTo(this.el, 'tl-tr', [2, 0]);

      this.helpImage.on("click", function()
      {
        helpWindow.setTitle("Automatic Refresh Interval Help");
        helpWindow.show();

        var textarea = helpWindow.getComponent("helpText_area");
        textarea.body.load({
          url:"/static/help/status-and-controls/automatic_refresh_help.html"});
      }, this);//scope
    },
    'change' : function(e, newVal, oldVal) {
      refreshGWStatusTask.interval = newVal*1000;
    }
  }
});

Paraxip.status.controls.setAutoStartGW = function(){
  var value = this.getValue();
  Ext.Ajax.request({
    url: Paraxip.baseUrl + "status/changeGWStartupType",
    method: 'POST',
    disableCaching: true,
    params: {
      'checkBoxValue':Ext.util.JSON.encode(value)
    }
  });
};

Paraxip.status.controls.setFailureAutoStartGW = function(){
  var value = this.getValue();
  Ext.Ajax.request({
    url: Paraxip.baseUrl + "status/changeGWFailureStartupType",
    method: 'POST',
    disableCaching: true,
    params: {
      'checkBoxValue':Ext.util.JSON.encode(value)
    }
  });
};

Paraxip.status.controls.statusButtonsPanel = new Ext.form.FormPanel({
  id          : 'Paraxip.status.controls.statusButtonsPanel'
  ,title      : 'Gateway Status'
  ,colspan    : 1
  ,width      : 486
  ,height     : 250
  ,labelWidth : 200
  ,bodyStyle  : "padding: 10px;text-align:left;"
  ,items      : [
    Paraxip.status.controls.autoRefreshTime,
  {
    xtype: 'label',
    html: '<html><br/></html>'
  },
    Paraxip.status.controls.refreshGWStatus
  ]
});

Paraxip.status.controls.autoStartGW = new Ext.form.Checkbox({
  id          : 'Paraxip.status.controls.autoStartGW'
  ,hideLabel  : true
  ,boxLabel   : "Automatically start gateway at startup"
  ,autoWidth  : true
  ,disabled   : true
  ,listeners  : {
    'check':Paraxip.status.controls.setAutoStartGW,
    'render':function(){
      if( !isGatewayInCLI ) {
        this.enable();
      }
      Ext.Ajax.request({
        url:Paraxip.baseUrl + "status/isGWAutomaticStartup",
        success: function(result){
          var response = Ext.util.JSON.decode(result.responseText);
          Paraxip.status.controls.autoStartGW.setValue(response.value);
        }
      });
    }
  }
});

Paraxip.status.controls.failureAutoStartGW = new Ext.form.Checkbox({
  id          : 'Paraxip.status.controls.failureAutoStartGW'
  ,boxLabel   : "Automatically restart gateway in case of failure"
  ,hideLabel  : true
  ,autoWidth  : true
  ,disabled   : true
  ,listeners  : {
    'check':Paraxip.status.controls.setFailureAutoStartGW,
    'render':function(){
      if( !isGatewayInCLI ) {
        this.enable();
      }
      Ext.Ajax.request({
        url:Paraxip.baseUrl + "status/isGWServiceFailureAutomaticRestartSet",
        success: function(result){
          var response = Ext.util.JSON.decode(result.responseText);
          Paraxip.status.controls.failureAutoStartGW.setValue(response.value);
        }
      });
    }
  }
});

Paraxip.status.controls.controlsButtonsPanel = new Ext.form.FormPanel({
  id        : 'Paraxip.status.controls.controlsButtonsPanel',
  title     : 'Gateway Controls',
  colspan   : 1,
  width     : 486,
  height    : 250,
  bodyStyle : "padding: 10px;text-align:left;",
  items     : [
    Paraxip.status.controls.startGwButton
    ,{
      xtype: 'label'
      ,html: '<html><br/></html>'
    }
    ,Paraxip.status.controls.stopGwButton
    ,{
      xtype: 'label'
      ,html: '<html><br/></html>'
    }
    ,Paraxip.status.controls.restartGwButton
    ,{
      xtype: 'label'
      ,html: '<html><br/></html>'
    }
    ,Paraxip.status.controls.autoStartGW
    ,{
      xtype: 'label'
      ,html: '<html><br/></html>'
    }
    ,Paraxip.status.controls.failureAutoStartGW
  ]
});

Paraxip.status.controls.messagePanel = new Ext.Panel({
  id        : 'Paraxip.status.controls.messagePanel'
  ,colspan  : 2
  ,width    : 972
  ,height   : 236
  ,autoLoad : {url:Paraxip.baseUrl + "static/gateway_down.html", nocache: false}
});

Paraxip.status.controls.controlsPanel = new Ext.Panel({
  id            : 'Paraxip.status.controls.controlsPanel',
  layout        : 'table',
  width         : 971,
  height        : 486,
  border        : false,
  layoutConfig  : {
    columns: 2
  },
  items         : [
    Paraxip.status.controls.controlsButtonsPanel,
    Paraxip.status.controls.statusButtonsPanel,
    Paraxip.status.controls.messagePanel
  ],
  listeners: {
    "render" : function() {
      // In lightbox mode, if pingGW is set to true, start immediately the
      // pingGW task.
      if( pingGW ) {
        Paraxip.changeGwStatusBar(
          Paraxip.initializingApplicationStatus, true, true);
        Paraxip.showProgressBar(
          "Service started. Initializing Gateway Application...");
        Paraxip.disableItems(true);
        Ext.TaskMgr.start(Paraxip.status.controls.pingGwTask);
      } 
      refreshGWStatusTask = {
        run: function() {
          Paraxip.isGwStartedCall();
        },
        interval: gwStatusRefreshTime*1000
      };
      Ext.TaskMgr.start(refreshGWStatusTask);
      Ext.Ajax.request({
        url: Paraxip.baseUrl + "getGCParameterValue",
        disableCaching: true,
        params: {"paramName":'netborder.oamWebServer.gwStatusRefreshTimeSec'},
        success: function(result){
          var response = Ext.util.JSON.decode(result.responseText);
          if (response.paramValue) {
            gwStatusRefreshTime = response.paramValue;
            if( refreshGWStatusTask != undefined ) {
              refreshGWStatusTask.interval = gwStatusRefreshTime*1000;
            }
            if( Paraxip.status.controls.autoRefreshTime != undefined ) {
              Paraxip.status.controls.autoRefreshTime.setValue(gwStatusRefreshTime);
            }
          }
        }
      });
    }
  }
});


