//-----------------------------------------------
// Call Routing form
//-----------------------------------------------
Ext.namespace("Paraxip.wizard");

var labelWidth = 285;

var call_routing_form = new Ext.Panel({
  title           : "Call Routing Configuration"
  ,layout         : 'form'
  ,autoWidth      : true
  ,autoHeigth     : true
  ,autoScroll     : true
  ,frame          : true
  ,header         : false
  ,id             : "Paraxip.wizard.call_routing_form"
  ,labelWidth     : labelWidth
  ,bodyStyle      : "text-align:left;padding-left:15px;"
  ,defaults       : {
    anchor: '65%'
  }
  ,items: [{}]
});

var CallRoutingDescriptionLabel = new Ext.form.Label({
   id      : "Paraxip.wizard.CallRoutingDescriptionLabel"
  ,text   : "Set your call routing information"
  ,style  : 'font-size:120%;font-weight:bold; line-height:60px;'
});

Paraxip.wizard.outboundMask = new Ext.form.TextField({
   id         : 'Paraxip.wizard.outboundMask'
  ,fieldLabel : "PSTN Outbound Number Mask"
  ,name       : "outbound_mask"
  ,helpUrl    : "quick-setup/qs-pstnOutboundMask.html"
  ,value      : '([\\+]?[-\\*\\#\\d]+)'
});

Paraxip.wizard.outboundPrefix = new Ext.form.TextField({
  id          : 'Paraxip.wizard.outboundPrefix'
  ,fieldLabel : "PSTN Outbound Number Prefix"
  ,name       : "outbound_prefix"
  ,helpUrl    : "quick-setup/qs-pstnOutboundPrefix.html"
  ,value      : ''
});


Paraxip.wizard.countryCode = new Ext.form.TextField({
   id         : 'Paraxip.wizard.countryCode'
  ,fieldLabel : "Country Code"
  ,name       : "country_code"
  ,helpUrl    : "quick-setup/qs-countryCode.html"
  ,value      : '1'
  ,maskRe     : /[0-9]/
});

Paraxip.wizard.pstnRoutingSet =  new Ext.form.FieldSet({
  id              : "Paraxip.wizard.pstnRoutingSet"
  ,name           : "pstnRouting"
  ,title          : 'PSTN Routing'
  ,autoHeight     : true
  ,collapsed      : false
  ,defaults       : {anchor: '-20'}
  ,items          : [
    Paraxip.wizard.outboundMask
    ,Paraxip.wizard.outboundPrefix
    ,Paraxip.wizard.countryCode
  ]
});


function setFailoverDisabled(bEnableFailover)
{
  Paraxip.wizard.failoverSipOutAddress.setDisabled(bEnableFailover);
  Paraxip.wizard.failoverSipOutPort.setDisabled(bEnableFailover);
  Paraxip.wizard.failoverSipTimeout.setDisabled(bEnableFailover);
}

var validateHostName = function(field)
{
  if (!isValidHostName(field.getValue())) {
    field.markInvalid(
      "Not a valid IPv4 address or a fully qualified domain name (FQDN).");
  } else {
    field.clearInvalid();
  }
}

Paraxip.wizard.outTransportProtocol = new Ext.form.ComboBox({
  id              : "Paraxip.wizard.outTransportProtocol"
  ,name           : "outTransportProtocol"
  ,fieldLabel     : "SIP Transport for outgoing calls"
  ,store          : ['udp','tcp', 'tls']
  ,forceSelection : true
  ,triggerAction  : 'all'
  ,mode           : 'local'
  ,typeAhead      : true
  ,value          : 'udp'
  ,selectOnFocus  : true
  ,allowBlank     : false
  ,helpUrl        : "quick-setup/qs-sip-transport.html"
});

//url input text
Paraxip.wizard.sipOutAddress = new Ext.form.TextField({
  id          : 'Paraxip.wizard.sipOutAddress'
  ,name       : "sipOutAddress"
  ,fieldLabel : "Primary SIP Server Address"
  ,style      : ".x-form-text {margin: 0px 0;top:0px;}"
  ,value      : 'localhost'
  ,helpUrl    : "quick-setup/qs-sip-primary-server.html"
  ,allowBlank : false
  ,listeners  : {
    'blur': validateHostName
  }
});

Paraxip.wizard.sipOutPort = new Ext.form.NumberField({
  id          : 'Paraxip.wizard.sipOutPort'
  ,name       : "sipOutPort"
  ,style      : ";.x-form-text {margin: 0px 0;top:0px;}"
  ,fieldLabel : "Primary SIP Server Port"
  ,allowBlank : false
});

Paraxip.wizard.sipRoutingSet =  new Ext.form.FieldSet({
  id              : "Paraxip.wizard.sipRoutingSet"
  ,name           : "sipRouting"
  ,title          : 'SIP Routing'
  ,autoHeight     : true
  ,collapsed      : false
  ,defaults       : {anchor: '-20'}
  ,items          : [
    Paraxip.wizard.outTransportProtocol
    ,Paraxip.wizard.sipOutAddress
    ,Paraxip.wizard.sipOutPort
  ]
});

Paraxip.wizard.failoverSipOutAddress = new Ext.form.TextField({
  id          : 'Paraxip.wizard.failoverSipOutAddress'
  ,name       : "failoverSipOutAddress"
  ,fieldLabel : "Backup SIP Server Address"
  ,value      : ''
  ,disabled   : true
  ,listeners  : {
    'blur': validateHostName
  }
});

Paraxip.wizard.failoverSipOutPort = new Ext.form.NumberField({
  id          : 'Paraxip.wizard.failoverSipOutPort'
  ,name       : "failoverSipOutPort"
  ,fieldLabel : "Backup SIP Server Port"
  ,disabled   : true
});

Paraxip.wizard.failoverSipTimeout = new Ext.form.NumberField({
  id          : 'Paraxip.wizard.failoverSipTimeout'
  ,name       : "failoverSipTimeout"
  ,value      : '5'
  ,disabled   : true
  ,fieldLabel : "Failover timeout (seconds)"
});

Paraxip.wizard.enableFailover =  new Ext.form.FieldSet({
  id              : "Paraxip.wizard.sipFailover"
  ,name           : "sipFailover"
  ,checkboxToggle : true
  ,title          : 'SIP Failover'
  ,autoHeight     : true
  ,collapsed      : true
  ,defaults       : {anchor: '-20'}
  ,helpUrl        : "quick-setup/qs-sip-backup-server.html"
  ,items          : [
    Paraxip.wizard.failoverSipOutAddress
    ,Paraxip.wizard.failoverSipOutPort
    ,Paraxip.wizard.failoverSipTimeout
  ]
  ,listeners  :
  {
    "expand" : function() {
      setFailoverDisabled(false);
      Paraxip.wizard.failoverSipOutAddress.focus(false, 1);
    }
    ,"collapse" : function() {
      setFailoverDisabled(true);
    }

  }
});

call_routing_form.add(CallRoutingDescriptionLabel);
call_routing_form.add(Paraxip.wizard.sipRoutingSet);
call_routing_form.add(Paraxip.wizard.enableFailover);
call_routing_form.add(Paraxip.wizard.pstnRoutingSet);
